-------------------------------------------------------------------------------
--                L  O  C  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

local addonName, L = ...;

local SPEC_DRUID_FERAL    = 103;
local SPEC_DRUID_GUARDIAN = 104;

local SPELL_DRUID_TRAVEL     =    783;
local SPELL_DRUID_CAT        =    768;
local SPELL_DRUID_BEAR       =   5487;
local SPELL_SHAMAN_GHOST     =   2645;
local SPELL_WORGEN_WILD      =  87840;
local SPELL_ITEM_HERB        =  88715;
local SPELL_GARRISON_ABILITY = 161691;
local SPELL_MANASABER        = 230987;
local SPELL_MASQUERADE       = 213213;

local ITEM_PET_BANDAGE    = 86143;
local ITEM_PET_SAFARI_HAT = 92738;

local MAPID_NORTHREND_WINTERGRASP = 123;
local MAPID_DRAENOR_NAGRAND       = 550;
local MAPID_BROKENISLES_SURAMAR   = 680;

local cityMapIds =
{
  [1161] = true, -- Boralus
  [ 125] = true, -- Dalaran, Northrend
  [ 126] = true, -- Dalaran, Northrend
  [ 627] = true, -- Dalaran, Broken Isles
  [ 628] = true, -- Dalaran, Broken Isles
  [  89] = true, -- Darnassus
  [  87] = true, -- Ironforge
  [  85] = true, -- Orgrimmar
  [  86] = true, -- Orgrimmar
  [ 111] = true, -- Shattrath
  [ 393] = true, -- Shrine of Seven Stars
  [ 394] = true, -- Shrine of Seven Stars
  [ 391] = true, -- Shrine of Two Moons
  [ 392] = true, -- Shrine of Two Moons
  [ 110] = true, -- Silvermoon City
  [ 622] = true, -- Stormshield
  [  84] = true, -- Stormwind
  [ 103] = true, -- The Exodar
  [ 775] = true, -- The Exodar
  [  88] = true, -- Thunder Bluff
  [ 624] = true, -- Warspear
  [  90] = true, -- Undercity
  [ 862] = true, -- Zuldazar
  [1181] = true, -- Zuldazar
  [1193] = true, -- Zuldazar
};

-- Used to build a mask for summoning pets
-- NOTE: The names must match the values returned by 'GetInstanceInfo()'. The
--       continent names are added at run-time. The IDs below must NEVER be
--       changed.
local petAreas =
{
  ["city"]     = {id = -1},
  ["party"]    = {id = -2},
  ["raid"]     = {id = -3},
  ["arena"]    = {id = -4},
  ["pvp"]      = {id = -5},
  ["scenario"] = {id = -6},
};

local mountEnv =
{
  {label = "Ground", action = "Run",  flag = 0x01},
  {label = "Air",    action = "Fly",  flag = 0x02},
  {label = "Water",  action = "Swim", flag = 0x04},
};

-- NOTE: The localized names of the spells are added to this table at run-time
local mountSpells =
{
  -- Travel Form
  [SPELL_DRUID_TRAVEL] =
  {
    flags = 0x07
  },
  -- Travel Form
  [SPELL_DRUID_CAT] =
  {
    flags = 0x01
  },
  -- Ghost Wolf
  [SPELL_SHAMAN_GHOST] =
  {
    flags = 0x01
  },
  -- Running Wild
  [SPELL_WORGEN_WILD] =
  {
    flags = 0x01
  },
  -- Herbouflage
  [SPELL_ITEM_HERB] =
  {
    flags = 0x00
  },
};

local mountHolidays =
{
  [L.flagHolidayHallowsEnd] =
  {
    [37011] = {skill = 1, flags = 0x03}, -- Magic Broom
    [33176] = {skill = 4, flags = 0x03}  -- Flying Broom
  },
};

local mountRestricted =
{
  [75207] = -- Vashj'ir Seahorse
  {
    [204] = true, -- Vashj'ir, Abyssal Depths
    [201] = true, -- Vashj'ir, Kelp'thar forest
    [205] = true, -- Vashj'ir Shimmering Expanse
  },
  [25953] = -- Blue Qiraji Battle Tank
  {
    [319] = true, -- Ahn`Qiraj, The Hive Underground
    [320] = true, -- Ahn`Qiraj, The Temple Gates
    [321] = true, -- Ahn`Qiraj, Vault of C'Thun
  },
  [26056] = -- Green Qiraji Battle Tank
  {
    [319] = true, -- Ahn`Qiraj, The Hive Underground
    [320] = true, -- Ahn`Qiraj, The Temple Gates
    [321] = true, -- Ahn`Qiraj, Vault of C'Thun
  },
  [26054] = -- Red Qiraji Battle Tank
  {
    [319] = true, -- Ahn`Qiraj, The Hive Underground
    [320] = true, -- Ahn`Qiraj, The Temple Gates
    [321] = true, -- Ahn`Qiraj, Vault of C'Thun
  },
  [26055] = -- Yellow Qiraji Battle Tank
  {
    [319] = true, -- Ahn`Qiraj, The Hive Underground
    [320] = true, -- Ahn`Qiraj, The Temple Gates
    [321] = true, -- Ahn`Qiraj, Vault of C'Thun
  },
};

local mapRestricted = {};

local mountTypeFlags =
{
  [230] = 0x05, -- Most ground mounts
  [231] = 0x05, -- Riding Turtle and Sea Turtle
  [232] = 0x04, -- Vashj'ir Seahorse
  [241] = 0x05, -- Blue, Green, Red & Yellow Qiraji Battle Tank
  [242] = 0x07, -- Swift Spectral Gryphon
  [247] = 0x07, -- Red Flying Cloud
  [248] = 0x07, -- Most flying mounts
  [254] = 0x04, -- Subdued Seahorse
  [269] = 0x05, -- Azure and Crimson Water Strider
};

local mountIdealFlags =
{
  [75614] = 0x03, -- Celestial Steed
  [48025] = 0x03, -- Headless Horseman's Mount
  [72286] = 0x03, -- Invincible
  [64731] = 0x04, -- Sea Turtle
};

local mountFlagSettings =
{
  [LE_MOUNT_JOURNAL_FILTER_COLLECTED]     = true,
  [LE_MOUNT_JOURNAL_FILTER_NOT_COLLECTED] = false
};

local mountZoneFunc = {};

local ridingInfo =
{
  33388, -- Apprentice Riding
  33391, -- Journeyman Riding
  34090, -- Artisan Riding
  34091, -- Expert Riding
  90265  -- Master Riding
};

local petFlagSettings =
{
  [LE_PET_JOURNAL_FILTER_COLLECTED]     = true,
  [LE_PET_JOURNAL_FILTER_NOT_COLLECTED] = false
};

-- Pets used as part of a quest
local petQuest =
{
  [31880] = -- Blood Elf Orphan Whistle
  {
      0,
  },
  [31881] = -- Draenei Orphan Whistle
  {
      0,
  },
  [30803] = -- Felhound Whistle
  {
    100 -- Hellfire Peninsula
  },
  [36865] = -- Golem Control Unit
  {
    116 -- Grizzly Hills
  },
  [18598] = -- Human Orphan Whistle
  {
      0,
  },
  [46397] = -- Oracle Orphan Whistle
  {
      0,
  },
  [18597] = -- Orcish Orphan Whistle
  {
      0,
  },
  [34253] = -- Sizzling Ember
  {
    100 --Hellfire Peninsula
  },
  [46362] =
  {
     81, -- Silithus
     71, -- Tanaris
     78, -- Un'Goro Crater
  },
  [68646] = -- Winterspring Cub
  {
     83 -- Winterspring
  },
  [46396] = -- Wolvar Orphan Whistle
  {
      0,
  },
  [31815] = -- Zeppit's Crystal
  {
    109, -- Netherstorm
  }
};

local petHoliday =
{
  [15698] = L.flagHolidayWinterVeil, -- Father Winter's Helper
  [15705] = L.flagHolidayWinterVeil, -- Winter's Little Helper
  [15710] = L.flagHolidayWinterVeil, -- Tiny Snowman
  [73741] = L.flagHolidayWinterVeil, -- Rotten Little Helper
};

local petInfo = {[0] = 0};
local petFilters =
{
  flags   = {},
  types   = {},
  sources = {},
  search  = ""
};

-- Used to save the tooltip for the last pet viewed, to save regenerating it
local lastPetInfo =
{
  text      = "",
  speciesId = 0,
  colour    = NORMAL_FONT_COLOR
};

local restrictedZones = {};
local continentalFlight = {};
local mountMacroText = "";
local ridingSkill = 0;
local hasClassRaceMounts = false;
local classRaceButtons = {};
local mountJournalToggles = {};
local mountFilteredBar;
local mountPriority;
local petJournalToggles  = {};
local petSummonedId = -1;
local petTimerRunning = false;
local petFilteredBar;
local petBandageButton;
local petJournalExtraButtons;

local mountSequence = {[0] = 0};

-------------------------------------------------------------------------------
--              G  L  O  B  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

NIGGLES_PETAREASEXCLUDED_DEFAULT =
{
  [petAreas.raid.id ] = true,
  [petAreas.arena.id] = true,
  [petAreas.pvp.id  ] = true,
};

StaticPopupDialogs["NIGGLES_COMPANIONS_CLEARALL"] =
{
  text           = "%s",
  button1        = OKAY,
  button2        = CANCEL,
  whileDead      = 1,
  OnAccept       = function(self, data)
    NigglesOptionsCompanionsClearAll(data);
  end,
  showAlert      = 1,
  timeout        = 0,
  exclusive      = 1,
  hideOnEscape   = 1,
  preferredIndex = STATICPOPUP_NUMDIALOGS,
};

StaticPopupDialogs["NIGGLES_COMPANIONS_SELECTALL"] =
{
  text           = "%s",
  button1        = OKAY,
  button2        = CANCEL,
  whileDead      = 1,
  OnAccept       = function(self, data)
    NigglesOptionsCompanionsSelectAll(data);
  end,
  showAlert      = 1,
  timeout        = 0,
  exclusive      = 1,
  hideOnEscape   = 1,
  preferredIndex = STATICPOPUP_NUMDIALOGS,
};

-------------------------------------------------------------------------------
--             L  O  C  A  L     D  E  F  I  N  I  T  I  O  N  S
-------------------------------------------------------------------------------

local getPetSpeciesOwnedInfo;
local mountPriorityManasaber;
local petBandageButtonCooldownUpdate;
local petBandageButtonOnEnter;
local petBandageButtonOnEvent;
local petBandageButtonOnHide;
local petBandageButtonOnShow;
local petBandageButtonUsabilityUpdate;
local petOnTimer;
local petSafariHatButtonCooldownUpdate;
local petSafariHatButtonOnEnter;
local petSafariHatButtonOnEvent;
local petSafariHatButtonOnHide;
local petSafariHatButtonOnShow;
local petSafariHatButtonUsabilityUpdate;
local petsListUpdate;
local unitTooltipSetOwned;

-------------------------------------------------------------------------------
--                L  O  C  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to change the tooltip for a battle pet during a battle to indicate
-- the level and quality of pets that the player owns of the same species.
--
local function battlePetTooltipSetOwned(self, text)
  -- Local Variables
  local speciesId;
  local parent = self:GetParent();

  -- Check the pet belongs to the enemy
  if ((NigglesDBC.settings.UiPetsOwned) and
      (parent.petOwner == LE_BATTLE_PET_ENEMY)) then
    -- Get the species of the enemy's battle pet
    speciesId = C_PetBattles.GetPetSpeciesID(LE_BATTLE_PET_ENEMY,
      parent.petIndex);

    -- Get the info about the pets owned of the specified species
    getPetSpeciesOwnedInfo(speciesId);

    -- Update the tooltip
    if (lastPetInfo.text ~= "") then
      self:SetFormattedText(lastPetInfo.text);
      self:SetTextColor(lastPetInfo.colour.r, lastPetInfo.colour.g,
        lastPetInfo.colour.b, 1);
    end
  end

  return;
end

--
-- Function to get info about the pets owned of a specified species. The info
-- is stored in a table to reduce memory usage.
--
getPetSpeciesOwnedInfo = function(speciesId)
  -- Check if the last pet info was obtained for is for the same species
  if (speciesId ~= lastPetInfo.speciesId) then
    -- Local Variables
    local chopMax = petInfo[0];
    local chopMid;
    local chopMin = 1;
    local isColourBlind = (GetCVar("colorblindMode") == "1" and true or false);
    local maxAllowed;
    local numCollected = 0;
    local speciesMask;
    local stringTable = {};

    -- Get the maximum number allowed for the species
    _, maxAllowed = C_PetJournal.GetNumCollectedInfo(speciesId);

    -- Search for the species in the list of pets
    speciesMask = speciesId*4096;
    while (chopMin < chopMax) do
      chopMid = math.floor((chopMax+chopMin)/2);
      if (bit.band(petInfo[chopMid], 0xFFFFF000) < speciesMask) then
        chopMin = chopMid+1;
      else
        chopMax = chopMid;
      end
    end

    -- Check a match was found for the species
    if ((chopMax == chopMin) and
        (bit.band(petInfo[chopMin], 0xFFFFF000) == speciesMask)) then
      speciesMask = speciesMask+0x00000FFF;
      while ((chopMax < petInfo[0]) and
             (petInfo[chopMax+1] < speciesMask)) do
        chopMax = chopMax+1;
      end
      numCollected = chopMax-chopMin+1;
    end

    -- Create the string for the species
    table.insert(stringTable, "Owned (");
    if (numCollected > 0) then
      for idx = chopMax, math.max(chopMax-maxAllowed+1, chopMin), -1 do
        quality = bit.band(petInfo[idx], 0x0000000F);
        level   = bit.rshift(bit.band(petInfo[idx], 0x00000FF0), 4);
        if (isColourBlind) then
          table.insert(stringTable, format(
            (idx < chopMax and "/%d%s" or "%d%s"),
            level, string.sub(L["PetQualityColourBlind"], quality, quality)));
        else
          table.insert(stringTable, format(
            (idx < chopMax and "/%s%d|r" or "%s%d|r"),
            ITEM_QUALITY_COLORS[quality-1].hex, level));
        end
      end
    end
    for idx = numCollected+1, maxAllowed do
      table.insert(stringTable, (idx > 1 and "/-" or "-"));
    end
    if (numCollected > maxAllowed) then
      table.insert(stringTable, "/...");
    end
    table.insert(stringTable, ")");
    lastPetInfo.speciesId = speciesId;
    lastPetInfo.text      = table.concat(stringTable);
    lastPetInfo.colour    =
      (numCollected < maxAllowed and NORMAL_FONT_COLOR or RED_FONT_COLOR);
  end

  return;
end

--
-- Function for processing clicks on a Mount toggle button
--
local function mountJournalToggleOnClick(self, button)
  -- Check the toggle is enabled and is being left-clicked.
  if ((self:IsEnabled()) and (button == "LeftButton")) then
    -- Update the state for the mount
    local spellId = self:GetParent().spellID;
    NigglesDBC.mounts[spellId] = bit.bxor((NigglesDBC.mounts[spellId] or 0),
      mountEnv[self:GetID()].flag);
    if (NigglesDBC.mounts[spellId] == 0) then
      NigglesDBC.mounts[spellId] = nil;
    end
  end

  return;
end

--
-- Function to display the tooltip for a Mount toggle button
--
local function mountJournalToggleOnEnter(toggle)
  -- Local Variables
  local toggleId = toggle:GetID();
  local actionColour;
  local textColour;

  -- Lock the highlighting of the toggle's parent
  toggle:GetParent():LockHighlight();

  -- Check the toggle's ID is valid
  if ((toggleId ~= nil) and
      (toggleId >= 1) and
      (toggleId <= #mountEnv)) then
    -- Work out what colours to use
    if (toggle:IsEnabled()) then
      actionColour = HIGHLIGHT_FONT_COLOR;
      textColour   = NORMAL_FONT_COLOR;
    else
      actionColour = GRAY_FONT_COLOR;
      textColour   = GRAY_FONT_COLOR;
    end

    -- Set the game tooltip
    GameTooltip:SetOwner(toggle, "ANCHOR_RIGHT");
    GameTooltip:SetMinimumWidth(100);
    GameTooltip:SetText(L["Mount"..mountEnv[toggleId].action],
      actionColour.r, actionColour.g, actionColour.b, 1, 1);
    GameTooltip:AddLine(L["Mount"..mountEnv[toggleId].action.."Tooltip"],
      textColour.r, textColour.g, textColour.b, true);
    GameTooltip:Show();
  else
    -- Hide the game tooltip
    GameTooltip:Hide();
  end

  return;
end

--
-- Function to process a click for the Mount Journal filtered bar
--
local function mountsFilteredBarOnClick(self, button)
  -- Check the filtered bar isn't 'hidden'
  if (self:GetHeight() > 1) then
    -- Clear all the filters
    for flag, _ in pairs(mountFlagSettings) do
      C_MountJournal.SetCollectedFilterSetting(flag, true);
    end
    C_MountJournal.SetAllSourceFilters(true);
    if (MountJournalSearchBox ~= nil) then
      MountJournalSearchBox:SetText("");
      MountJournalSearchBox:ClearFocus();
    end
  end

  -- 'Hide' the filtered bar
  self:SetHeight(1);

  return;
end

--
-- Function called when a registered event occurs for the Mount Journal
-- filtered bar.
--
local function mountsFilteredBarOnEvent(self, event, ...)
  -- Process the event
  if (event == "MOUNT_JOURNAL_SEARCH_UPDATED") then
    -- Local Variables
    local _;
    local isFiltered = false;
    local numSources = C_PetJournal.GetNumPetSources();

    -- Search for any filters
    for flag, _ in pairs(mountFlagSettings) do
      if (not C_MountJournal.GetCollectedFilterSetting(flag)) then
        isFiltered = true;
        break;
      end
    end
    if (not isFiltered) then
      for idx = 1, numSources do
        if ((C_MountJournal.IsValidSourceFilter(idx)) and
            (not C_MountJournal.IsSourceChecked(idx))) then
          isFiltered = true;
          break;
        end
      end
    end
    if ((not isFiltered) and (MountJournalSearchBox:GetText() ~= "")) then
      isFiltered = true;
    end

    -- Show/Hide the filtered bar by adjusting its height, which
    -- automatically adjusts the position of the scroll frame
    if (isFiltered) then
      mountFilteredBar:SetFormattedText(L["FilteredShown"],
        C_MountJournal.GetNumDisplayedMounts());
    end
    mountFilteredBar:SetHeight(isFiltered and 18 or 1);
  end

  return;
end

--
-- Function called when the Mount Journal filtered bar is shown
--
local function mountsFilteredBarOnHide(self)
  -- Unregister for updates to the Mount Journal's list
  self:UnregisterEvent("MOUNT_JOURNAL_SEARCH_UPDATED");

  return;
end

--
-- Function called when the Mount Journal filtered bar is hidden
--
local function mountsFilteredBarOnShow(self)
  -- Register for updates to the Mount Journal's list
  self:RegisterEvent("MOUNT_JOURNAL_SEARCH_UPDATED");

  -- Update the Pet Journal filtered bar
  mountsFilteredBarOnEvent(self, "MOUNT_JOURNAL_SEARCH_UPDATED");

  return;
end

--
-- Function to update the Mount Journal scroll frame
--
local function mountsListUpdate()
  -- Check if the scroll frame needs updating
  if ((MountJournal.ListScrollFrame:IsShown()) and
      (NigglesDBC.settings.MountsEnabled == true) and
      (next(mountJournalToggles))) then
    -- Local Variables
    local button;
    local buttons;
    local isCollected;
    local mountFaction;
    local mountFlags;
    local mountId;
    local mountSpellId;
    local mountType;
    local mountInfo = NigglesDBC.mounts;
    local numButtons;
    local numEnvironments = #mountEnv;
    local playerFaction;
    local scrollFrame = MountJournal.ListScrollFrame;
    local toggle;

    -- Initialise some variables;
    buttons       = MountJournal.ListScrollFrame.buttons;
    numButtons    = #scrollFrame.buttons;
    playerFaction = (UnitFactionGroup("player") == "Alliance" and 1 or 0);

    -- Update all the buttons in the Mounts scroll frame
    for buttonIdx = 1, numButtons do
      -- Check the current button is displaying a mount
      button = buttons[buttonIdx];
      if ((button.spellID or 0) > 0) then
        -- Get the info for the mount the button is for
        _, mountSpellId, _, _, _, _, _, _, mountFaction, _, isCollected,
          mountId = C_MountJournal.GetDisplayedMountInfo(button.index);
        mountType   =
          select(5, C_MountJournal.GetMountInfoExtraByID(mountId));
        mountFlags  = (mountTypeFlags[mountType] or 0x07);

        -- Check the player is allowed to use the mount
        if ((mountFaction ~= nil) and
            (mountFaction ~= playerFaction) and
            (mountInfo[button.spellID] ~= nil)) then
          mountInfo[button.spellID] = nil;
        end

        -- Set the enabled and checked state of the button's toggles
        for toggleIdx = 1, numEnvironments do
          toggle = mountJournalToggles[buttonIdx][toggleIdx];
          if ((isCollected) and
              ((mountFaction == nil) or (mountFaction == playerFaction)) and
              (bit.band(mountFlags, mountEnv[toggleIdx].flag) ~= 0)) then
            toggle:SetEnabled(true);
            toggle:SetChecked(
              (mountInfo[button.spellID] ~= nil) and
              (bit.band(mountInfo[button.spellID],
                 mountEnv[toggleIdx].flag) ~= 0));
          else
            toggle:SetEnabled(false);
            toggle:SetChecked(false);
          end
        end
      else
        button:Hide();
      end
    end
  end

  return;
end

--
-- Function to deal with special rules for mounting in Suramar in Broken Isles
--
mountPriorityManasaber = function(playerMapId)
  -- Local Variables
  local priority = 0;

  -- Work out if the Arcanist's Manasaber should be given priority
  if ((playerMapId == MAPID_BROKENISLES_SURAMAR) and
      (((ExtraActionButton1:IsVisible()) and
        (ExtraActionButton1.action == 169)) or
        (L.playerAuras.has(SPELL_MASQUERADE)))) then
    priority = 3;
  end

  return priority;
end

--
-- Function to deal with special rules for mounting in Nagrand on Draenor
--
mountZoneFunc[MAPID_DRAENOR_NAGRAND] = function(isModKeyDown, isSwimming)
  -- Local Variables
  local mountText = nil;
  local specId;

  -- Check if the garrison ability should be used
  if ((NigglesDBC.settings.MountsNagrandTalbuk) and
      (IsUsableSpell(SPELL_GARRISON_ABILITY)) and
      (GetSpellTexture(SPELL_GARRISON_ABILITY) ==
        "Interface\\Icons\\ability_mount_goatmount") and
      ((not continentalFlight[L.continents.player()]) or
       (isModKeyDown))) then
    -- Check if the character is a druid
    if (Niggles.playerClass == "DRUID") then
      -- Use the spec's main form as a mount
      specId = GetSpecializationInfo(GetSpecialization());
      if (specId == SPEC_DRUID_FERAL) then
        mountText = "/use "..GetSpellInfo(SPELL_DRUID_CAT);
      elseif (specId == SPEC_DRUID_GUARDIAN) then
        mountText = "/use "..GetSpellInfo(SPELL_DRUID_BEAR);
      end
    end

    -- Use the Corral's Telaari Talbuk if no other 'mount' is more suitable
    if (mountText == nil) then
      mountText = "/use "..L["Telaari Talbuk"];
    end
  end

  return nil, mountText;
end

--
-- Function to deal with special rules for mounting in Wintergrasp on Northrend
--
mountZoneFunc[MAPID_NORTHREND_WINTERGRASP] = function(isModKeyDown, isSwimming)
  -- Local Variables
  local flagIdx = 1;
  local timeToBattle;
  local mountFlags = {};

  -- Work out if the player can fly
  _, _, _, _, timeToBattle = GetWorldPVPAreaInfo(1);
  if ((isSwimming) and (not isModKeyDown)) then
    mountFlags[flagIdx] = 0x04;
    flagIdx = flagIdx+1;
  end
  if (((isSwimming) or (not isModKeyDown)) and
      (continentalFlight[L.continents.player()]) and
      (timeToBattle > 0)) then
    mountFlags[flagIdx] = 0x02;
    flagIdx = flagIdx+1;
  end
  mountFlags[flagIdx] = 0x01;

  return mountFlags, nil
end

--
-- Function to initialise the info for pet areas
--
local function petAreasInit()
  -- Local Variables
  local idx;
  local info;
  local mask;
  local name;

  -- Initialise the pet masks, using localized names for the continents
  for name, info in pairs(petAreas) do
    info.name = name;
    info.mask = bit.rshift(0x80000000, -(info.id+1));
  end
  for idx = 1, L.continents.length do
    id, name, mask = L.continents.byIndex(idx);
    petAreas[name] =
    {
      id   = id,
      name = name,
      mask = mask,
    };
  end

  return;
end
  
--
-- Function to update the cooldown frame for the Pet Bandage button
--
petBandageButtonCooldownUpdate = function(self)
  -- Local Variables
  local start, duration, enable = GetItemCooldown(ITEM_PET_BANDAGE);

  -- Update the cooldown
  CooldownFrame_Set(self.cooldown, start, duration, enable);

  return;
end

--
-- Function called when the mouse enters the Pet Bandage button
--
petBandageButtonOnEnter = function(self)
  -- Set the tooltip
  GameTooltip:SetOwner(self, "ANCHOR_RIGHT");
  GameTooltip:SetItemByID(ITEM_PET_BANDAGE);

  -- Add any explanation for the button being disabled
  if (not C_PetJournal.IsJournalUnlocked()) then
    GameTooltip:AddLine(L["PetBandageLocked"], RED_FONT_COLOR.r,
      RED_FONT_COLOR.g, RED_FONT_COLOR.b, true);
  elseif (C_PetBattles.IsInBattle()) then
    GameTooltip:AddLine(L["PetBandageInBattle"], RED_FONT_COLOR.r,
      RED_FONT_COLOR.g, RED_FONT_COLOR.b, true);
  end

  -- Show the tooltip;
  self.UpdateTooltip = petBandageButtonOnEnter;
  GameTooltip:Show();

  return;
end

--
-- Function to process events for the Pet Bandage button
--
petBandageButtonOnEvent = function(self, event, ...)
  -- Process the event
  if (event == "BAG_UPDATE_COOLDOWN") then
    petBandageButtonCooldownUpdate(self);
    if (GameTooltip:GetOwner() == self) then
      petBandageButtonOnEnter(self);
    end
  elseif ((event == "BAG_UPDATE") or
          (event == "PET_BATTLE_OPENING_START") or
          (event == "PET_BATTLE_CLOSE")) then
    petBandageButtonUsabilityUpdate(self);
  end

  return;
end

--
-- Function called when the Pet Bandage button is hidden
--
petBandageButtonOnHide = function(self)
  -- Un-register for events
  self:UnregisterAllEvents();

  return;
end

--
-- Function called when the Pet Bandage button is shown. It registers the
-- button for some events that need monitoring.
--
petBandageButtonOnShow = function(self)
  -- Register for events
  self:RegisterEvent("BAG_UPDATE_COOLDOWN");
  self:RegisterEvent("BAG_UPDATE");
  self:RegisterEvent("PET_BATTLE_OPENING_START");
  self:RegisterEvent("PET_BATTLE_CLOSE");

  -- Refresh the button's state
  petBandageButtonCooldownUpdate(self);
  petBandageButtonUsabilityUpdate(self);

  return;
end

--
-- Function to update the usability of the Pet Bandage button
--
petBandageButtonUsabilityUpdate = function(self)
  -- Local Variables
  local isDisabled = false;
  local isLocked = false;
  local itemCount = GetItemCount(ITEM_PET_BANDAGE, false, false);

  -- Work out the correct state for the button
  if ((not C_PetJournal.IsJournalUnlocked()) or
      (C_PetBattles.IsInBattle()) or
      ((itemCount > 0) and (not IsUsableItem(ITEM_PET_BANDAGE)))) then
    isLocked   = true;
    isDisabled = true;
  elseif (itemCount < 1) then
    isDisabled = true;
  end

  -- Update the state of the button
  self:SetButtonState("NORMAL", isDisabled);
  self.icon:SetDesaturated(isLocked);
  self.lockIcon:SetShown(isLocked);
  self.blackCover:SetShown(isDisabled);
  self.count:SetShown(not isLocked);

  -- Update the number of items in the character's bags
  self.count:SetText(itemCount > 1000 and "*" or itemCount);

  return;
end

--
-- Function to update the cooldown frame for the Pet Safari Hat button
--
petSafariHatButtonCooldownUpdate = function(self)
  -- Local Variables
  local start, duration, enable = GetItemCooldown(self.info.id);

  -- Update the cooldown
  CooldownFrame_Set(self.cooldown, start, duration, enable);

  return;
end

--
-- Function called when the mouse enters the Pet Safari Hat button
--
petSafariHatButtonOnEnter = function(self)
  -- Set the tooltip
  GameTooltip:SetOwner(self, "ANCHOR_RIGHT");
  GameTooltip:SetToyByItemID(ITEM_PET_SAFARI_HAT);

  -- Add any explanation for the button being disabled
  if (C_PetBattles.IsInBattle()) then
    GameTooltip:AddLine(L["PetSafariHatInBattle"], RED_FONT_COLOR.r,
      RED_FONT_COLOR.g, RED_FONT_COLOR.b, true);
  end

  -- Show the tooltip;
  self.UpdateTooltip = petSafariHatButtonOnEnter;
  GameTooltip:Show();

  return;
end

--
-- Function to process events for the Pet Safari Hat button
--
petSafariHatButtonOnEvent = function(self, event, ...)
  -- Process the event
  if ((event == "PET_BATTLE_OPENING_START") or
      (event == "PET_BATTLE_CLOSE")) then
    petSafariHatButtonUsabilityUpdate(self);
  elseif (event == "SPELL_UPDATE_COOLDOWN") then
    petSafariHatButtonCooldownUpdate(self);
  end

  return;
end

--
-- Function called when the Pet Safari Hat button is hidden
--
petSafariHatButtonOnHide = function(self)
  -- Un-register for events
  self:UnregisterAllEvents();

  return;
end

--
-- Function called when the Pet Safari Hat button is shown. It registers the
-- button for some events that need monitoring.
--
petSafariHatButtonOnShow = function(self)
  -- Register for events
  self:RegisterEvent("SPELL_UPDATE_COOLDOWN");
  self:RegisterEvent("PET_BATTLE_OPENING_START");
  self:RegisterEvent("PET_BATTLE_CLOSE");

  -- Refresh the button's state
  petSafariHatButtonCooldownUpdate(self);
  petSafariHatButtonUsabilityUpdate(self);

  return;
end

--
-- Function to update the usability of the Pet SafariHat button
--
petSafariHatButtonUsabilityUpdate = function(self)
  -- Local Variables
  local isDisabled = false;
  local isLocked = false;
  local hasToy = PlayerHasToy(ITEM_PET_SAFARI_HAT);

  -- Work out the correct state for the button
  if ((not C_PetJournal.IsJournalUnlocked()) or
      (C_PetBattles.IsInBattle()) or
      (not hasToy)) then
    isLocked   = true;
    isDisabled = true;
  elseif (not hasToy) then
    isDisabled = true;
  end

  -- Update the state of the button
  self:SetButtonState("NORMAL", isDisabled);
  self.icon:SetDesaturated(isLocked);
  self.lockIcon:SetShown(isLocked);
  self.blackCover:SetShown(isDisabled);

  return;
end

--
-- Function to check if pets can be summoned/dismissed
--
local function petsCanSummon()
  -- Local Variables
  local auras = L.playerAuras.get();

  -- Check if pets can be safely summoned/dismissed
  return ((not IsFlying()) and
          (not IsFalling()) and
          (not IsStealthed()) and
          (GetNumLootItems() == 0) and
          (not UnitInVehicle("player")) and
          (not UnitUsingVehicle("player")) and
          (not UnitIsDeadOrGhost("player")) and
          (not UnitIsCharmed("player")) and
          (not UnitIsPossessed("player")) and
          (UnitCastingInfo("player") == nil) and
          (UnitChannelInfo("player") == nil) and
          (auras[L["Food"]                ] ~= true) and
          (auras[L["Drink"]               ] ~= true) and
          (auras[L["Feign Death"]         ] ~= true) and
          (auras[L["Camouflage"]          ] ~= true) and
          (auras[L["Herbouflage"]         ] ~= true) and
          (auras[L["Spirit of Redemption"]] ~= true) and
          ((BarberShopFrame == nil) or
           (not BarberShopFrame:IsVisible())));
end

--
-- Function to process a click for the Pet Journal filtered bar
--
local function petsFilteredBarOnClick(self, button)
  -- Check the filtered bar isn't 'hidden'
  if (self:GetHeight() > 1) then
    -- Clear all the filters
    for flag, _ in pairs(petFlagSettings) do
      C_PetJournal.SetFilterChecked(flag, true);
    end
    C_PetJournal.SetAllPetTypesChecked(true);
    C_PetJournal.SetAllPetSourcesChecked(true);
    C_PetJournal.ClearSearchFilter();
    if (PetJournalSearchBox ~= nil) then
      PetJournalSearchBox:SetText("");
      PetJournalSearchBox:ClearFocus();
    end
  end

  -- 'Hide' the filtered bar
  self:SetHeight(1);

  return;
end

--
-- Function to process events for the Pet Journal filtered bar
--
local function petsFilteredBarOnEvent(self, event, ...)
  -- Process the event
  if (event == "PET_JOURNAL_LIST_UPDATE") then
    if ((PetJournal:IsShown()) and
        (NigglesDBC.settings.UiCompanionsFiltered)) then
      -- Local Variables
      local isFiltered = false;

      -- Search for any filters
      for flag, value in pairs(petFlagSettings) do
        if (not C_PetJournal.IsFilterChecked(flag)) then
          isFiltered = true;
          break;
        end
      end
      if (not isFiltered) then
        for idx = 1, C_PetJournal.GetNumPetTypes() do
          if (not C_PetJournal.IsPetTypeChecked(idx)) then
            isFiltered = true;
            break;
          end
        end
      end
      if (not isFiltered) then
        for idx = 1, C_PetJournal.GetNumPetSources() do
          if (not C_PetJournal.IsPetSourceChecked(idx)) then
            isFiltered = true;
            break;
          end
        end
      end
     if ((not isFiltered) and
         (PetJournalSearchBox ~= nil) and
         (PetJournalSearchBox:GetText() ~= "")) then
       isFiltered = true;
     end

      -- Show/Hide the filtered bar by adjusting its height, which
      -- automatically adjusts the position of the scroll frame
      if (isFiltered) then
        self:SetFormattedText(L["FilteredShown"], C_PetJournal.GetNumPets());
      end
      self:SetHeight(isFiltered and 18 or 1);
    end
  end

  return;
end

--
-- Function called when the Pet Journal filtered bar is shown
--
local function petsFilteredBarOnHide(self)
  -- Unregister for updates to the Pet Journal's list
  self:UnregisterEvent("PET_JOURNAL_LIST_UPDATE");

  return;
end

--
-- Function called when the Pet Journal filtered bar is hidden
--
local function petsFilteredBarOnShow(self)
  -- Register for updates to the Pet Journal's list
  self:RegisterEvent("PET_JOURNAL_LIST_UPDATE");

  -- Update the Pet Journal filtered bar
  petsFilteredBarOnEvent(self, "PET_JOURNAL_LIST_UPDATE");

  return;
end

--
-- Function to clear the current pet filters, saving them for later restoration.
-- This function MUST be followed by a call to 'petFiltersRestore', preferably
-- before the calling function returns.
--
local function petsFiltersClear()
  -- Local Variables
  local cpj = C_PetJournal;

  -- Disable the processing of 'PET_JOURNAL_LIST_UPDATE' events to prevent
  -- an infinite loop
  if (PetJournal ~= nil) then
    PetJournal:UnregisterEvent("PET_JOURNAL_LIST_UPDATE")
  end
  Niggles:UnregisterEvent("PET_JOURNAL_LIST_UPDATE")

  -- Save the current filter settings and then set them to the required state
  for flag, value in pairs(petFlagSettings) do
    petFilters.flags[flag] = cpj.IsFilterChecked(flag);
    cpj.SetFilterChecked(flag, value);
  end
  for idx = 1, cpj.GetNumPetTypes() do
    petFilters.types[idx] = cpj.IsPetTypeChecked(idx);
  end
  for idx = 1, cpj.GetNumPetSources() do
    petFilters.sources[idx] = cpj.IsPetSourceChecked(idx);
  end
  cpj.SetAllPetTypesChecked(true);
  cpj.SetAllPetSourcesChecked(true);
  if (PetJournalSearchBox ~= nil) then
    petFilters.search = PetJournalSearchBox:GetText();
    if ((petFilters.search == SEARCH) or (petFilters.search == nil)) then
      petFilters.search = "";
    end
  else
    petFilters.search = "";
  end
  cpj.ClearSearchFilter();

  return;
end

--
-- Function to restore previously saved values for the pet filters.
--
local function petFiltersRestore()
  -- Local Variables
  local cpj = C_PetJournal;

  -- Restore the filtering settings
  for flag, value in pairs(petFilters.flags) do
    cpj.SetFilterChecked(flag, value);
  end
  for idx, value in ipairs(petFilters.types) do
    cpj.SetPetTypeFilter(idx, value);
  end
  for idx, value in ipairs(petFilters.sources) do
    cpj.SetPetSourceChecked(idx, value);
  end
  cpj.SetSearchFilter(petFilters.search);

  -- Restore the processing of 'PET_JOURNAL_LIST_UPDATE' events
  if (PetJournal ~= nil) then
    PetJournal:RegisterEvent("PET_JOURNAL_LIST_UPDATE")
  end
  Niggles:RegisterEvent("PET_JOURNAL_LIST_UPDATE")

  return;
end

--
-- Function called when the mouse pointer enters a button in the Pet Journal
-- list
--
local function petJournalButtonOnEnter(button)
  -- Local Variables
  local petId = button:GetParent().petID;

  -- Check the tooltip should be displayed and the button's pet ID was obtained
  if ((NigglesDBC.settings.UiPetJournalTooltip) and (petId)) then
    -- Get the pet's info
    local speciesId, _, level, _, _, _, _, name =
      C_PetJournal.GetPetInfoByPetID(petId)
    local _, maxHealth, power, speed, rarity =
      C_PetJournal.GetPetStats(petId);

    -- Position the Game tooltip for use by the Battle Pet tooltip
    GameTooltip:SetOwner(button, "ANCHOR_RIGHT");

    -- Display the Battle Pet tooltip
    BattlePetToolTip_Show(speciesId, level, rarity-1, maxHealth, power,
      speed, name);

    -- Save the button the tooltip is being displayed for
    petJournalListButton = button:GetParent();
  end

  return;
end

--
-- Function called when the mouse pointer leaves a button in the Pet Journal
-- list
--
local function petJournalButtonOnLeave(button)
  -- Hide the Battle Pet tooltip.
  BattlePetTooltip:Hide();

  -- Clear the button the tooltip is being displayed for
  petJournalListButton = nil;

  return;
end

--
-- Function called every few seconds to summon or dismiss a random pet, if
-- appropriate
--
petOnTimer = function()
  -- Local Variables
  local activeHolidays;
  local areasExcluded = NigglesDBC.settings.PetAreasExcluded;
  local creatureId;
  local isExcluded = false;
  local mapId = C_Map.GetBestMapForUnit("player");
  local petDismiss = false;
  local petId = C_PetJournal.GetSummonedPetGUID();
  local petSummon = false;
  local randomList = {{}, {}};
  local zoneType;

  -- Clear the flag that indicates the timer is running
  petTimerRunning = false;

  -- Check if pets have been disabled
  if ((not NigglesDBC.settings.PetsEnabled) or
      (next(NigglesDBC.pets) == nil)) then
    -- Simply return to cancel the timer
    return;
  end

  -- Check the character isn't in combat
  if (not UnitAffectingCombat("player")) then
    _, zoneType = GetInstanceInfo();
    if (zoneType == "none") then
      if ((cityMapIds[C_Map.GetBestMapForUnit("player")] ~= nil) and
          (areasExcluded[petAreas["city"].id] == true)) then
        isExcluded = true;
      else
        isExcluded = (areasExcluded[L.continents.player()] == true);
      end
    else
      isExcluded = (areasExcluded[petAreas[zoneType].id] == true);
    end

    -- Based on the current zone, adjust the flags
    petSummon  = ((not isExcluded) and (petId == nil));
    petDismiss = ((isExcluded) and (petId ~= nil) and (petId == petSummonedId));

    -- Check if any special pets might need to be summoned in the current zone
    if ((petSummon == true) and
        ((restrictedZones[0] == true) or
         (restrictedZones[mapId] == true))) then
      petSummon = false;
    end

    -- Check if a pet should be summoned or dismissed
    if ((petSummon == true) and
        (next(NigglesDBC.pets) ~= nil) and
        (petsCanSummon())) then
      -- Get the active holidays
      activeHolidays = L.holidaysGetActive();

      -- Work out which pets have been selected by the player
      for petId, creatureId in pairs(NigglesDBC.pets) do
        if (petHoliday[creatureId] ~= nil) then
          if (bit.band(activeHolidays, petHoliday[creatureId]) ~= 0) then
            table.insert(randomList[1], petId);
          end
        else
          table.insert(randomList[2], petId);
        end
      end

      -- Check if there are any holiday pets to summon
      if (next(randomList[1]) ~= nil) then
        petSummonedId = randomList[1][math.random(#randomList[1])];
      elseif (next(randomList[2]) ~= nil) then
        petSummonedId = randomList[2][math.random(#randomList[2])];
      else
        petSummonedId = nil;
      end

      -- Summon a pet, if possible
      if (petSummonedId ~= nil) then
        C_PetJournal.SummonPetByGUID(petSummonedId);
      end
    elseif ((petDismiss == true) and (petsCanSummon())) then
      -- Dismiss the pet, by summoning it again
      C_PetJournal.SummonPetByGUID(petSummonedId);
      petSummonedId = nil;
    end
  end

  -- Call this function again in a short while
  C_Timer.After(15.0, petOnTimer);
  petTimerRunning = true;

  return;
end

--
-- Function to update the Pet Journal scroll frame
--
petsListUpdate = function()
  -- Local Variables
  local scrollFrame = PetJournal.listScroll;
  local numButtons = #scrollFrame.buttons;
  local toggle;
  local button;

  -- Update all the buttons in the Pets scroll frame
  for idx = 1, numButtons do
    -- Get the current button
    button = scrollFrame.buttons[idx];

    -- Check the Pets option is enabled
    if (NigglesDBC.settings.PetsEnabled) then
      -- Set the enabled and checked state of the button's toggle
      toggle = petJournalToggles[idx];
      if ((button.dragButton:IsEnabled()) and
          (button.petID ~= nil)) then
        toggle:SetEnabled(true);
        if (NigglesDBC.pets[button.petID] ~= nil) then
          toggle:SetChecked(true);
        else
          toggle:SetChecked(false);
        end
      else
        toggle:SetEnabled(false);
        toggle:SetChecked(false);
      end
    end

    -- Update the tooltip if its being displayed for this button
    if (button == petJournalListButton) then
      petJournalButtonOnEnter(button.dragButton);
    end
  end

  return;
end

--
-- Function to (un)register for events monitored to aid in summoning pets
--
local function petsRegisterEvents()
  -- Register for the required events
  if (NigglesDBC.settings.UiPetsOwned) then
    Niggles:RegisterEvent("PET_BATTLE_LEVEL_CHANGED");
    Niggles:RegisterEvent("UPDATE_SUMMONPETS_ACTION");
  else
    Niggles:UnregisterEvent("PET_BATTLE_LEVEL_CHANGED");
    Niggles:UnregisterEvent("UPDATE_SUMMONPETS_ACTION");
  end
  if ((NigglesDBC.settings.UiPetsOwned) or
      ((NigglesDBC.settings.PetsEnabled) and
       (next(NigglesDBC.pets) ~= nil))) then
    Niggles:RegisterEvent("PET_JOURNAL_LIST_UPDATE");
  else
    Niggles:UnregisterEvent("PET_JOURNAL_LIST_UPDATE");
  end
  if ((NigglesDBC.settings.PetsEnabled) and
      (next(NigglesDBC.pets) ~= nil)) then
    Niggles:RegisterEvent("BAG_UPDATE");
  else
    Niggles:UnregisterEvent("BAG_UPDATE");
  end

  -- Create a timer for summoning pets, if required
  if ((NigglesDBC.settings.PetsEnabled) and
      (next(NigglesDBC.pets) ~= nil) and
      (not petTimerRunning)) then
    C_Timer.After(15.0, petOnTimer);
    petTimerRunning = true;
  end

  -- Create a timer to update the list of pets owned in a few seconds
  if (NigglesDBC.settings.UiPetsOwned) then
    C_Timer.After(1.0, NigglesPetsGetOwned);
  end

  -- Force the last pet info to be updated
  lastPetInfo.speciesId = 0;

  return;
end

--
-- Function for processing clicks on a Pet toggle button
--
local function petToggleOnClick(self, mouseButton)
  -- Local Variables
  local parent = self:GetParent();
  local wasEmpty = (next(NigglesDBC.pets) == nil);

  -- Check the toggle is enabled and is being left-clicked.
  if ((self:IsEnabled()) and
      (mouseButton == "LeftButton") and
      (parent ~= nil) and
      (parent.petID ~= nil)) then
    -- Update the state for the pet
    if (self:GetChecked()) then
      NigglesDBC.pets[parent.petID] =
        select(11, C_PetJournal.GetPetInfoByPetID(parent.petID));
    else
      NigglesDBC.pets[parent.petID] = nil;
    end

    -- Register for events, if required
    if (((next(NigglesDBC.pets) ~= nil) and (wasEmpty)) or
        ((next(NigglesDBC.pets) == nil) and (not wasEmpty))) then
      petsRegisterEvents();
    end
  end

  return;
end

--
-- Function to display the tooltip for a Pet toggle button
--
local function petToggleOnEnter(toggle)
  -- Local Variables
  local textColour;

  -- Lock the highlighting of the toggle's parent
  toggle:GetParent():LockHighlight();

  -- Work out what colour to use
  if (toggle:IsEnabled()) then
    textColour = NORMAL_FONT_COLOR;
  else
    textColour = GRAY_FONT_COLOR;
  end

  -- Set the game tooltip
  GameTooltip:SetOwner(toggle, "ANCHOR_RIGHT");
  GameTooltip:SetMinimumWidth(100);
  GameTooltip:SetText(L["PetTooltip"],
    textColour.r, textColour.g, textColour.b, 1, 1);
  GameTooltip:Show();

  return;
end

--
-- Function to change the tooltip for a battle pet to indicate the level and
-- quality of pets that the player owns of the same species.
--
local function petTooltipSetOwned(self, text)
  -- Local Variables
  local parent = self:GetParent();

  -- Check the enhanced battle pet tooltip should be displayed
  if ((NigglesDBC.settings.UiPetsOwned) and
      (parent.speciesID ~= nil) and
      (text ~= nil)) then
    -- Get the info about the pets owned of the specified species
    getPetSpeciesOwnedInfo(parent.speciesID);

    -- Update the tooltip
    if (lastPetInfo.text ~= "") then
      self:SetFormattedText(lastPetInfo.text);
      self:SetTextColor(lastPetInfo.colour.r, lastPetInfo.colour.g,
        lastPetInfo.colour.b, 1);
    end
  end

  return;
end

--
-- Function to change the tooltip for a battle pet to indicate the level and
-- quality of pets that the player owns of the same species.
--
unitTooltipSetOwned = function(self)
  -- Local Variables
  local _, unit = self:GetUnit();
  local numLines;
  local lineIdx = 0;
  local fontStringPrefix;
  local lineText;

  -- Check the enhanced battle pet tooltip should be displayed
  if ((NigglesDBC.settings.UiPetsOwned) and
      (unit ~= nil) and
      (UnitIsBattlePet(unit))) then
    -- Search for the line containing the details of the pets owned
    numLines         = self:NumLines();
    fontStringPrefix = self:GetName().."TextLeft";
    for idx = numLines, 1, -1 do
      lineText = _G[fontStringPrefix..idx]:GetText();
      if ((lineText == UNIT_CAPTURABLE) or
          (string.find(lineText, "^Collected %(%d/%d%)$") ~= nil)) then
        lineIdx = idx;
        break;
      end
    end

    -- Check a line containing the details of the pets owned was found
    if (lineIdx > 0) then
      -- Get the info about the pets owned of the specified species
      getPetSpeciesOwnedInfo(UnitBattlePetSpeciesID(unit));

      -- Update the tooltip
      _G[fontStringPrefix..lineIdx]:SetText(lastPetInfo.text);
    end
  end

  return;
end

-------------------------------------------------------------------------------
--                A  D  D  O  N     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to return a list of areas for pet summoning
--
function L.petAreasGetList()
  -- Local Variables
  local areaInfo;
  local list = {};

  -- Initialise the info for the pet areas, if required
  if (petAreas.city.mask == nil) then
    petAreasInit();
  end

  for _, areaInfo in pairs(petAreas) do
    table.insert(list, {id = areaInfo.id, name = areaInfo.name});
  end
  table.sort(list,
    function(first, second)
      if ((first.id < 0) and (second.id < 0)) then
        return (first.id > second.id);
      elseif (first.id < 0) then
        return false;
      elseif (second.id < 0) then
        return true;
      end
      return (first.name < second.name);
    end);

  return list;
end

-------------------------------------------------------------------------------
--               G  L  O  B  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to initialise the Niggles Mount and Pet functionality. This
-- function initialise those aspects of the features that do not require the
-- Mount & Pet UI to be loaded.
--
function NigglesCompanionsInit()
  -- Local Variables
  local spellId;
  local zones;

  -- Work out if the player can have class/race "mounts"
  if ((Niggles.playerClass == "DRUID") or
      (Niggles.playerClass == "SHAMAN") or
      (Niggles.playerRace  == "Worgen")) then
    hasClassRaceMounts = true;
  else
    hasClassRaceMounts = false;
    NigglesDBC.settings.MountsClassRace = false;
  end

  -- Initialise the pet masks, using localized names for the continents
  if (petAreas.city.mask == nil) then
    petAreasInit();
  end

  -- Initialise a list of zones with restricted mounts
  for spellId, zones in pairs(mountRestricted) do
    for zone, _ in pairs(zones) do
      if (mapRestricted[zone] == nil) then
        mapRestricted[zone] = {};
      end
      table.insert(mapRestricted[zone], spellId);
    end
  end

  -- Initialise the list of mount spells
  NigglesMountsSpellsUpdate();

  -- Initialise the list of restricted zones, based on items in bags
  NigglesPetsBagUpdate();

  -- Hook the tooltips used to display battle pet info
  hooksecurefunc(BattlePetTooltip.Owned, "SetText",
    petTooltipSetOwned);
  hooksecurefunc(FloatingBattlePetTooltip.Owned, "SetText",
    petTooltipSetOwned);
  hooksecurefunc(PetBattlePrimaryUnitTooltip.CollectedText, "SetText",
    battlePetTooltipSetOwned);
  GameTooltip:HookScript("OnTooltipSetUnit", unitTooltipSetOwned);

  return;
end

--
-- Function called when Blizzard's Pet Journal AddOn is loaded
--
function NigglesCompanionsOnLoad()
  -- Local Variables
  local buttons;
  local frame;

  -- Check character isn't in combat, since some actions can't be done then
  if (not InCombatLockdown()) then
    -- Check the Mount Journal scroll frame exists
    if (MountJournal.ListScrollFrame ~= nil) then
      -- Hook the function that updates the Pets list
      hooksecurefunc("MountJournal_UpdateMountList", mountsListUpdate);
      MountJournal.ListScrollFrame.update = MountJournal_UpdateMountList;
    end

    -- Check the Pets Journal scroll frame exists
    if (PetJournal.listScroll ~= nil) then
      -- Hook the function that updates the Pets list
      hooksecurefunc("PetJournal_UpdatePetList",
        petsListUpdate);
      PetJournal.listScroll.update = PetJournal_UpdatePetList;

      -- Add script handlers to Pets Journal scroll frame buttons
      buttons  = PetJournal.listScroll.buttons;
      for idx = 1, #buttons do
        buttons[idx].dragButton:SetScript("OnEnter", petJournalButtonOnEnter);
        buttons[idx].dragButton:SetScript("OnLeave", petJournalButtonOnLeave);
      end
    end

    -- Enable/disable the Mount & Pet features
    NigglesCompanionsEnable();
  else
    tinsert(Niggles.postCombat, NigglesCompanionsOnLoad);
  end

  -- Hook functions into the 'Prev' and 'Next' buttons on various frames
  frame = HeirloomsJournal.PagingFrame;
  frame.PrevPageButton:HookScript("OnMouseUp", NigglesFirstLastPage);
  frame.NextPageButton:HookScript("OnMouseUp", NigglesFirstLastPage);
  frame = WardrobeCollectionFrame.ItemsCollectionFrame.PagingFrame;
  frame.PrevPageButton:HookScript("OnMouseUp", NigglesFirstLastPage);
  frame.NextPageButton:HookScript("OnMouseUp", NigglesFirstLastPage);

  return;
end

--
-- Function to enable/disable the Mount & Pet Journal features
--
function NigglesCompanionsEnable()
  -- Local Variables
  local button;
  local buttons;
  local numButtons;
  local numEnvironments = #mountEnv;
  local prevButton;
  local settings = NigglesDBC.settings;
  local toggle;
  local toggles;

  -- Check character isn't in combat, since some frames can't be created then
  if (not InCombatLockdown()) then
    -- Check the Mount Journal exists
    if (MountJournal ~= nil) then
      -- Assign some info to more convenient variables
      buttons    = MountJournal.ListScrollFrame.buttons;
      numButtons = #(MountJournal.ListScrollFrame.buttons);

      -- Create, show or hide the toggles for each button
      for buttonIdx = 1, numButtons do
        -- Get the current button
        button = buttons[buttonIdx];

        -- Initialise the button's toggle table, if required
        if ((settings.MountsEnabled) and
            (mountJournalToggles[buttonIdx] == nil)) then
          mountJournalToggles[buttonIdx] = {};
        end

        -- Create, show or hide the toggles for each button
        toggles = mountJournalToggles[buttonIdx];
        if (toggles ~= nil) then
          for toggleIdx = 1, numEnvironments do
            -- Create the toggle, if required
            if ((settings.MountsEnabled) and (toggles[toggleIdx] == nil)) then
              toggle = CreateFrame("CheckButton", nil, button,
                "NigglesToggleTemplate");
              toggle:SetID(toggleIdx);
              toggle:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT",
                -(5+((numEnvironments-toggleIdx)*16)), 5);
              toggle:SetScript("OnClick", mountJournalToggleOnClick);
              toggle:SetScript("OnEnter", mountJournalToggleOnEnter);
              toggles[toggleIdx] = toggle;
            end

            -- Show/Hide the toggle, if required
            if (toggles[toggleIdx] ~= nil) then
              toggles[toggleIdx]:SetShown(settings.MountsEnabled);
            end
          end
        end
      end

      -- Update the Mount Journal scroll frame, if required
      if (MountJournal.ListScrollFrame:IsShown()) then
        mountsListUpdate();
      end

      -- Create the filtered bar for the Mount Journal, if required
      if ((settings.UiCompanionsFiltered) and
          (mountFilteredBar == nil)) then
        -- Create the filtered bar
        mountFilteredBar = CreateFrame("BUTTON", nil, MountJournal,
          "NigglesFilteredBarTemplate");
        mountFilteredBar:SetSize(250, 18);
        mountFilteredBar:SetPoint("TOPLEFT", MountJournalSearchBox,
          "BOTTOMLEFT", -9, -3);
        mountFilteredBar:SetScript("OnClick", mountsFilteredBarOnClick);
        mountFilteredBar:SetScript("OnEvent", mountsFilteredBarOnEvent);
        mountFilteredBar:SetScript("OnHide", mountsFilteredBarOnHide);
        mountFilteredBar:SetScript("OnShow", mountsFilteredBarOnShow);
        mountFilteredBar:Show();

        -- Adjust the position of Mount Journal scroll frame and its scroll bar
        MountJournalListScrollFrameScrollBar:SetPoint("TOPLEFT",
          MountJournal.LeftInset, "TOPRIGHT", 2, -16);
        MountJournalListScrollFrame:SetPoint("TOPLEFT", mountFilteredBar,
          "BOTTOMLEFT", -3, -3);
      end

      -- Update the Mount Journal scroll frame, if required
      if (MountJournal.ListScrollFrame:IsShown()) then
        mountsListUpdate();
      end

      -- Hide the Pet filtered bar, if required
      if ((mountFilteredBar ~= nil) and
          (not settings.UiCompanionsFiltered)) then
        mountFilteredBar:SetHeight(1);
      end
    end

    -- Check the Pet Journal exists
    if (PetJournal ~= nil) then
      -- Initialise some variables
      buttons    = PetJournal.listScroll.buttons;
      numButtons = #(PetJournal.listScroll.buttons);

      -- Create, show or hide the toggle for each button
      for buttonIdx = 1, numButtons do
        -- Get the current button
        button = buttons[buttonIdx];

        -- Create the toggle, if required
        if ((settings.PetsEnabled) and
            (petJournalToggles[buttonIdx] == nil)) then
          toggle = CreateFrame("CheckButton", nil, button,
            "NigglesToggleTemplate");
          toggle:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", -5, 5);
          toggle:SetScript("OnClick", petToggleOnClick);
          toggle:SetScript("OnEnter", petToggleOnEnter);
          petJournalToggles[buttonIdx] = toggle;
        end

        -- Show/Hide the toggle, if required
        if (petJournalToggles[buttonIdx] ~= nil) then
          petJournalToggles[buttonIdx]:SetShown(settings.PetsEnabled);
        end
      end

      -- Update the Pet Journal scroll frame, if required
      if ((PetJournal.listScroll:IsShown()) and (settings.PetsEnabled)) then
        petsListUpdate();
      end

      -- Create the filtered bar for the Pet Journal, if required
      if ((settings.UiCompanionsFiltered) and (petFilteredBar == nil)) then
        -- Create the filtered bar
        petFilteredBar = CreateFrame("BUTTON", nil, PetJournal,
          "NigglesFilteredBarTemplate");
        petFilteredBar:SetSize(250, 18);
        petFilteredBar:SetPoint("TOPLEFT", PetJournalSearchBox,
          "BOTTOMLEFT", -9, -3);
        petFilteredBar:SetScript("OnClick", petsFilteredBarOnClick);
        petFilteredBar:SetScript("OnEvent", petsFilteredBarOnEvent);
        petFilteredBar:SetScript("OnHide", petsFilteredBarOnHide);
        petFilteredBar:SetScript("OnShow", petsFilteredBarOnShow);
        petFilteredBar:Show();

        -- Adjust the position of Pet Journal scroll frame and its scroll bar
        PetJournalListScrollFrameScrollBar:SetPoint("TOPLEFT",
          PetJournalLeftInset, "TOPRIGHT", 2, -16);
        PetJournalListScrollFrame:SetPoint("TOPLEFT", petFilteredBar,
          "BOTTOMLEFT", -3, -3);
      end

      -- Update the Pet Journal scroll frame, if required
      if (PetJournal.listScroll:IsShown()) then
        petsListUpdate();
      end

      -- Hide the Pet filtered bar, if required
      if ((petFilteredBar ~= nil) and (not settings.UiCompanionsFiltered)) then
        petFilteredBar:SetHeight(1);
      end

      -- Show/hide the extra Pet Journal buttons
      for _, buttonInfo in ipairs(petJournalExtraButtons) do
        -- Create the button, if required
        if ((settings[buttonInfo.setting]) and (buttonInfo.button == nil)) then
          button = CreateFrame("BUTTON", nil, PetJournal,
            "NigglesPetButtonTemplate");
          button.info = buttonInfo;
          button.icon:SetTexture(GetItemIcon(buttonInfo.id));
          button:RegisterForClicks("LeftButtonUp", "RightButtonUp");
          button:RegisterForDrag("LeftButton", "RightButton");
          button:SetScript("OnEnter", buttonInfo.onEnter);
          button:SetScript("OnEvent", buttonInfo.onEvent);
          button:SetScript("OnHide",  buttonInfo.onHide);
          button:SetScript("OnShow",  buttonInfo.onShow);
          button:SetAttribute("type", buttonInfo.type);
          button:SetAttribute(buttonInfo.type,
            buttonInfo.typePrefix..buttonInfo.id);
          buttonInfo.button = button;
        else
          button = buttonInfo.button;
        end

        -- Show/hide the button
        if (settings[buttonInfo.setting]) then
          button:SetPoint("TOPRIGHT", (prevButton or PetJournalHealPetButton),
            "TOPLEFT", -5, 0);
          button:Show();
          prevButton = button;
        elseif (button ~= nil) then
          button:Hide();
        end
      end
      PetJournalHealPetButtonSpellName:SetShown(prevButton == nil);
      PetJournalSummonRandomFavoritePetButtonSpellName:SetShown(
        prevButton == nil);
    end

    -- Enable/disable pet summoning
    petsRegisterEvents();
  else
    tinsert(Niggles.postCombat, NigglesCompanionsEnable);
  end

  return;
end

--
-- Function reset the macro of the hidden mount button after it has been
-- 'clicked'.
--
function NigglesMountPostClick(button)
  -- Clear the button's macro
  if (not InCombatLockdown()) then
    button:SetAttribute("macrotext", mountMacroText);
  end

  return
end

--
-- Function to work out what to do when a macro calls '/click NigglesMount'.
-- This function is called before the button is clicked. It works out what
-- to do based on the player's current status and changes the macro of the
-- hidden button accordingly.
--
function NigglesMountPreClick(button)
  -- Local Variables
  local auras = L.playerAuras.get();
  local activeAura = "";
  local activeHolidays = 0;
  local envIdx;
  local flagIdx = 1;
  local inCombat = InCombatLockdown();
  local isModKeyDown = IsModifierKeyDown();
  local isSwimming = IsSwimming();
  local macroText;
  local mountFlags = nil;
  local mountSeqIdx;
  local priority;
  local randomInfo = nil;
  local randomList = {};
  local randomPriority;
  local spellName = nil;
  local playerMapId = C_Map.GetBestMapForUnit("player");

  -- Get the active holidays, if required
  if (NigglesDBC.settings.MountsHoliday) then
    activeHolidays = L.holidaysGetActive();
  end

  -- Check if any of the mount spells are active
  for _, spellInfo in pairs(mountSpells) do
    if (auras[spellInfo.name] ~= nil) then
      activeAura = spellInfo.name;
    end
  end

  -- Work out what to do, based on the player's current location and state
  if (UnitInVehicle("player")) then
    -- Exit any vehicle the player is in
    VehicleExit();
  elseif (IsMounted()) then
    -- Dismount from any mount the player is on
    Dismount();
  elseif ((activeAura ~= "") and (not inCombat)) then
    -- Cancel the aura
    button:SetAttribute("macrotext", "/cancelaura "..activeAura);
  else
    -- Call the mount function for the zone, if required
    if ((not inCombat) and
        (type(mountZoneFunc[playerMapId]) == "function")) then
      mountFlags, macroText = mountZoneFunc[playerMapId](isModKeyDown,
        isSwimming);
      if (macroText ~= nil) then
        button:SetAttribute("macrotext", macroText);
        mountFlags = nil;
      end
    end

    -- Check if the mount function returned info
    if ((mountFlags == nil) and (macroText == nil)) then
      mountFlags = {};
      if ((isSwimming) and (not isModKeyDown)) then
        mountFlags[flagIdx] = 0x04;
        flagIdx = flagIdx+1;
      end
      -- TODO: Check the logic of this
      if (((isSwimming) or (not isModKeyDown)) and
          ((continentalFlight[L.continents.player()]) or
           (IsInInstance())) and
          (IsFlyableArea())) then
        mountFlags[flagIdx] = 0x02;
        flagIdx = flagIdx+1;
      end
      mountFlags[flagIdx] = 0x01;
    end
  end

  -- Check if the player should be mounted
  if (mountFlags ~= nil) then
    if (inCombat) then
      UIErrorsFrame:AddMessage(SPELL_FAILED_AFFECTING_COMBAT,
        1.0, 0.1, 0.1, 1.0);
    else
      -- Search each specified mount type for a mount
      for _, currentFlag in ipairs(mountFlags) do
        -- Work out the index for the current environment
        envIdx = (currentFlag == 0x04 and 3 or currentFlag);

        -- Check if holiday mounts should be used
        if (NigglesDBC.settings.MountsHoliday == true) then
          -- Initialise some variables
          randomInfo = GetItemInfo;

          -- Search for selected holiday mounts with the current flag
          for holidayFlag, holidayMounts in pairs(mountHolidays) do
            -- Check if the holiday is active
            if (bit.band(activeHolidays, holidayFlag) ~= 0) then
              -- Search for usable holiday mounts
              for itemId, itemInfo in pairs(holidayMounts) do
                if ((itemInfo.skill <= ridingSkill) and
                    (bit.band(itemInfo.flags, currentFlag) ~= 0) and
                    (IsUsableItem(itemId))) then
                  table.insert(randomList, itemId);
                end
              end
            end
          end
        end

        -- Check if there are any mounts restricted to the player's map
        if ((next(randomList) == nil) and
            (mapRestricted[playerMapId] ~= nil)) then
          -- Initialise some variables
          randomInfo     = GetSpellInfo;
          randomPriority = 0;

          -- Search for selected restricted mounts with the current flag
          for _, spellId in ipairs(mapRestricted[playerMapId]) do
            if ((NigglesDBC.mounts[spellId] ~= nil) and
                (bit.band(NigglesDBC.mounts[spellId], currentFlag) ~= 0) and
                (IsUsableSpell(spellId))) then
              -- Work out the mount's priority
              if (mountPriority[spellId] ~= nil) then
                if (type(mountPriority[spellId][envIdx]) == "function") then
                  priority = mountPriority[spellId][envIdx](playerMapId);
                else
                  priority = mountPriority[spellId][envIdx];
                end
              elseif (mountSequence[spellId] ~= nil) then
                priority = 2+((mountSequence[0]-mountSequence[spellId])/10);
              else
                priority = 2.9;
              end

              -- If the mount has a higher priority than the random list...
              if (priority > randomPriority) then
                -- ...reset the random list
                wipe(randomList);
                randomPriority = priority
              end

              -- Add the mount to the random list if it is high enough priority
              if (priority >= randomPriority) then
                table.insert(randomList, spellId);
              end
            end
          end
        end

        -- Search for unrestricted mounts that have the current flag
        if (next(randomList) == nil) then
          -- Initialise some variables
          randomInfo     = GetSpellInfo;
          randomPriority = 0;

          -- Search for selected mounts with the current flag
          for spellId, spellFlags in pairs(NigglesDBC.mounts) do
            if ((bit.band(spellFlags, currentFlag) ~= 0) and
                (mountRestricted[spellId] == nil) and
                (IsUsableSpell(spellId)) and
                -- Note: This is required because 'IsUsableSpell()' returns
                --       true for Travel Form even when it isn't
                ((spellId ~= SPELL_DRUID_TRAVEL) or
                 (IsOutdoors()) or
                 (IsSwimming()))) then
              -- Work out the mount's priority
              if (mountPriority[spellId] ~= nil) then
                if (type(mountPriority[spellId][envIdx]) == "function") then
                  priority = mountPriority[spellId][envIdx](playerMapId);
                else
                  priority = mountPriority[spellId][envIdx];
                end
              elseif (mountSequence[spellId] ~= nil) then
                priority = 2+((mountSequence[0]-mountSequence[spellId])/10);
              else
                priority = 2.9;
              end

              -- If the mount has a higher priority than the random list...
              if (priority > randomPriority) then
                -- ...reset the random list
                wipe(randomList);
                randomPriority = priority
              end

              -- Add the mount to the random list if it is high enough priority
              if (priority >= randomPriority) then
                table.insert(randomList, spellId);
              end
            end
          end
        end

        -- Check if any suitable mounts were found
        if (next(randomList) ~= nil) then
          -- Choose a random mount
          randomId = randomList[math.random(#(randomList))];

          -- Clear the mount from the mount sequence list, if required
          if (mountSequence[randomId] ~= nil) then
            mountSequence[(mountSequence[randomId]%3)+1] = nil;
          end

          -- Add the mount to the mount sequence list
          mountSeqIdx = (mountSequence[0]%3)+1;
          if (mountSequence[mountSeqIdx] ~= nil) then
            mountSequence[mountSequence[mountSeqIdx]] = nil;
          end
          mountSequence[mountSeqIdx] = randomId;
          mountSequence[randomId]    = mountSequence[0];
          mountSequence[0]           = mountSequence[0]+1;

          -- Choose a random mount
          spellName = randomInfo(randomId);
          break;
        end
      end

      -- Update the button with the mount's spell name
      if (spellName ~= nil) then
        button:SetAttribute("macrotext", mountMacroText.."/cast "..spellName);
      else
        button:SetAttribute("macrotext", "");
        UIErrorsFrame:AddMessage(L["No suitable mount"], 1.0, 0.1, 0.1, 1.0);
      end
    end
  end

  return;
end

--
-- Function called when the player's spells change
--
function NigglesMountsSpellsUpdate()
  -- Local Variables
  local id;
  local numContinents = L.continents.length;
  local spellId;

  -- Work out the player's riding skill
  ridingSkill = 0;
  for idx, spellId in ipairs(ridingInfo) do
    if (IsSpellKnown(spellId)) then
      ridingSkill = idx;
    end
  end

  -- Initialise the continental flight info
  for idx = 1, numContinents do
    id, _, _, spellId = L.continents.byIndex(idx);
    continentalFlight[id] = 
      ((ridingSkill >= 3) and
       (spellId >= 0) and
       ((spellId == 0) or (IsSpellKnown(spellId))));
  end

  -- Check if the player has the skill to fly
  if (ridingSkill >= 3) then
    -- Initialise the continents with no flight restrictions
    continentalFlight[1] = true; -- Kalimdor
    continentalFlight[2] = true; -- Eastern Kingdoms
    continentalFlight[3] = true; -- Outland
    continentalFlight[4] = true; -- Northrend
    continentalFlight[5] = true; -- The Maelstrom
    continentalFlight[6] = true; -- Pandaria

    -- Check if the player can fly in Draenor
    if (IsSpellKnown(191645)) then -- Draenor Pathfinder
      continentalFlight[7] = true;
    end
   
    -- Check if the player can fly in the Broken Isles
    if (IsSpellKnown(233368)) then -- Broken Isles Pathfinder, Rank 2
      continentalFlight[8] = true;
    end
  end

  -- Get names of the mount spells & build a string to cancel their auras
  mountMacroText = "";
  for spellId, spellInfo in pairs(mountSpells) do
    spellInfo.name = GetSpellInfo(spellId);
    if ((spellInfo.name ~= nil) and
        ((spellInfo.flags == nil) or
         (IsSpellKnown(spellId)))) then
      mountMacroText = mountMacroText.."/cancelaura [combat] "..
        spellInfo.name.."\n";
    end
  end

  return
end

--
-- Function to clear all selected companions/mounts
--
function NigglesOptionsCompanionsClearAll(button)
  -- Check the specified type is valid
  if ((button.companionType == "mounts") or
      (button.companionType == "pets")) then
    -- Clear all the selected companions
    wipe(Niggles.updatedOptions[button.companionType]);

    -- Enable/disable options based on the new settings
    NigglesOptionsEnable(button:GetParent());
  end

  return;
end

--
-- Function to enable/disable the 'Class/Racial' check buttons in the Niggles
-- Options Companions panel
--
function NigglesOptionsMountsClassRaceEnable(object, setting)
  -- Local Variables
  local button;
  local buttonIdx = 1;
  local order =
  {
    SPELL_DRUID_TRAVEL,
    SPELL_DRUID_CAT,
    SPELL_SHAMAN_GHOST,
    SPELL_WORGEN_WILD,
  };
  local parent = object:GetParent();
  local prevButton;
  local spellIcon;
  local spellInfo;

  -- Enable/disable the appropriate objects
  if (setting == "MountsClassRace") then
    -- Enable/disable the 'Class/Racial' check button
    object:SetEnabled((Niggles.updatedOptions.settings.MountsEnabled) and
      (hasClassRaceMounts));
  else
    -- Initialise all the class/racial check buttons to the correct state
    -- NOTE: This is done here rather than an initialisation routine to cater
    --       for characters learning new abilities
    for _, spellId in ipairs(order) do
      -- Check if a button should be displayed for the spell
      spellInfo = mountSpells[spellId];
      if ((spellInfo ~= nil) and (spellInfo.flags ~= nil)) then
        -- Create the button, if required
        button = classRaceButtons[buttonIdx];
        if (button == nil) then
          if (buttonIdx > 1) then
            button = CreateFrame("checkbutton", nil, parent,
              "NigglesClassRaceButtonTemplate");
            button:SetPoint("TOPLEFT", prevButton, "TOPRIGHT", 3, 0);
          else
            button = object;
          end
          classRaceButtons[buttonIdx] = button;
        end

        -- Check the button exists
        if (button ~= nil) then
          -- Update the button's icon and state
          button.spellId    = spellId;
          button.spellFlags = spellInfo.flags;
          _, _, spellIcon = GetSpellInfo(spellId);
          button:SetNormalTexture(spellIcon);
          if ((IsSpellKnown(spellId)) and
              (Niggles.updatedOptions.settings.MountsEnabled   == true) and
              (Niggles.updatedOptions.settings.MountsClassRace == true)) then
            button:Enable();
            button:SetChecked(Niggles.updatedOptions.mounts[spellId]);
          else
            button:Disable();
            button:SetChecked(false);
          end

          -- Increment the button index
          buttonIdx = buttonIdx+1;

          -- Save the button's handle
          prevButton = button;
        end
      end
    end
  end

  return;
end

--
-- Function to select all companions/mounts as selected
--
function NigglesOptionsCompanionsSelectAll(button)
  -- Local Variables
  local _;
  local spellFlags;
  local spellId;
  local numCompanions;
  local isFactionSpecific;
  local faction;
  local hideOnChar;
  local isCollected;
  local mountType;

  -- Check the specified type is valid
  if (button.companionType == "mounts") then
    -- Clear all selected mounts
    wipe(Niggles.updatedOptions.mounts);

    -- Select all know mounts
    numCompanions = C_MountJournal.GetNumMounts();
    for idx = 1, numCompanions do
       _, spellId, _, _, _, _, _, isFactionSpecific, faction, hideOnChar,
         isCollected = C_MountJournal.GetDisplayedMountInfo(idx);
      if ((spellId ~= nil) and
          ((not isFactionSpecific) or
           (faction == Niggles.factionId)) and
          (isCollected) and
          (not hideOnChar)) then
        -- Check if the mount has ideal flags specified for it
        if (mountIdealFlags[spellId] ~= nil) then
          -- Use the ideal flags
          if (mountIdealFlags[spellId] ~= 0) then
            Niggles.updatedOptions.mounts[spellId] = mountIdealFlags[spellId];
          end
        else
          -- Work out the spell flags for the mount
          _, _, _, _, mountType =
            C_MountJournal.DisplayedGetMountInfoExtra(idx);
          spellFlags = (mountTypeFlags[mountType] or 0x07);

          -- Set the mount's flags to select its optimal environment
          if (bit.band(spellFlags, 0x02) ~= 0) then
            Niggles.updatedOptions.mounts[spellId] = 0x02;
          elseif (bit.band(spellFlags, 0x01) ~= 0) then
            Niggles.updatedOptions.mounts[spellId] = 0x01;
          elseif (bit.band(spellFlags, 0x04) ~= 0) then
            Niggles.updatedOptions.mounts[spellId] = 0x04;
          end
        end
      end
    end

    -- Select all known mount spells
    for spellId, spellInfo in pairs(mountSpells) do
      if (IsSpellKnown(spellId)) then
        Niggles.updatedOptions.mounts[spellId] = spellInfo.flags;
      end
    end

    -- Enable/disable options based on the new settings
    NigglesOptionsEnable(button:GetParent());
  elseif (button.companionType == "pets") then
    -- Clear all selected pets
    wipe(Niggles.updatedOptions.pets);

    -- Clear the current filter settings
    petsFiltersClear();

    -- Select all know pets
    numCompanions = C_PetJournal.GetNumPets();
    for idx = 1, numCompanions do
      petId, _, _, _, _, _, _, _, _, _, creatureId =
        C_PetJournal.GetPetInfoByIndex(idx);
      if (petId ~= nil) then
        Niggles.updatedOptions.pets[petId] = creatureId;
      end
    end

    -- Restore the filtering settings
    petFiltersRestore();
  end

  return;
end

--
-- Function called when the contents of the player's bags changes. It searches
-- for any quest items used to summon pets
--
function NigglesPetsBagUpdate()
  -- Local Variables
  local itemId;
  local itemMapIds;
  local mapId

  -- Clear the list of restricted zones
  wipe(restrictedZones);

  -- Search for any items used to summon quest pets
  for itemId, itemMapIds in pairs(petQuest) do
    if (GetItemCount(itemId, false, false) > 0) then
      for _, mapId in ipairs(itemMapIds) do
        restrictedZones[mapId] = true;
      end
    end
  end

  return;
end

--
-- Function to refresh the list of owned pets, even if the number hasn't
-- changed.
--
function NigglesPetRefreshOwned()
  -- Set the number of pets in the list to 0 to force a refresh
  petInfo[0] = 0;

  -- Create a timer to update the list of pets owned in a few seconds
  C_Timer.After(1.0, NigglesPetsGetOwned);

  return
end

--
-- Function to get info about the pets owned
--
function NigglesPetsGetOwned()
  -- Local Variables
  local infoMissing = false;
  local _, numPets = C_PetJournal.GetNumPets();
  local petId;
  local petLevel;
  local petQuality;
  local petSpeciesId;

  -- Check if the number of pets has changed since the last call
  if (numPets ~= petInfo[0]) then
    -- Check all the selected pets are still known
    if (NigglesDBC.settings.PetsEnabled) then
      for petId, _ in pairs(NigglesDBC.pets) do
        if ((string.find(petId, "BattlePet%-0%-") ~= 1) or
            (C_PetJournal.GetPetInfoByPetID(petId) == nil)) then
          NigglesDBC.pets[petId] = nil;
        end
      end
    end

    -- Check if enhanced Battle Pet tooltips are to be displayed
    if (NigglesDBC.settings.UiPetsOwned) then
      -- Clear the current filter settings
      petsFiltersClear();

      -- Get the species, level and quality info for all know pets
      petInfo[0] = numPets;
      for idx = 1, numPets do
        petId, petSpeciesId, _, _, petLevel =
          C_PetJournal.GetPetInfoByIndex(idx);
        if (petId ~= nil) then
          _, _, _, _, petQuality = C_PetJournal.GetPetStats(petId);
          if (petQuality ~= nil) then
            petInfo[idx] = (petSpeciesId*4096)+(petLevel*16)+petQuality;
          end
        else
          infoMissing = true;
          break;
        end
      end

      -- Check if info for any pets was missing
      if (infoMissing) then
        -- Scan again shortly
        NigglesPetRefreshOwned()
      else
        -- Remove any deleted pets from the table
        for idx = #petInfo, numPets+1, -1 do
          petInfo[idx] = nil;
        end

        -- Sort the pets
        table.sort(petInfo);
      end

      -- Restore the filtering settings
      petFiltersRestore();

      -- Force the last pet info to be updated
      lastPetInfo.speciesId = 0;

      -- Free the large amount of memory used to get the pet info
      collectgarbage("collect");
    end
  end

  -- Check for summoned pets, if required
  if (petSummonedId == -1) then
    -- Check if any summoned pet is selected
    petId = C_PetJournal.GetSummonedPetGUID();
    if ((petId ~= nil) and (NigglesDBC.pets[petId] ~= nil)) then
      petSummonedId = petId;
    else
      petSummonedId = nil;
    end
  end

  return;
end

-- Function to get the flag for a pet area
function NigglesPetsGetAreaFlag(areaType)
  return petAreas[areaType];
end

--
-- Function to update the info for a pet when it levels
--
function NigglesPetUpdateLevel(petSlot, petNewLvl)
  -- Local Variables
  local _;
  local chopMax = petInfo[0];
  local chopMid;
  local chopMin = 1;
  local petId;
  local petSpeciesId;
  local key;
  local petOldLvl;

  -- Get the pet's ID
  petId = C_PetJournal.GetPetLoadOutInfo(petSlot);
  if (petId ~= nil) then
    -- Get the pet's species and quality
    petSpeciesId, _, petOldLvl = C_PetJournal.GetPetInfoByPetID(petId);
    _, _, _, _, petQuality = C_PetJournal.GetPetStats(petId);

    -- Search for the pet
    key = (petSpeciesId*4096)+(petOldLvl*16)+petQuality;
    while (chopMin < chopMax) do
      chopMid = math.floor((chopMax+chopMin)/2);
      if (petInfo[chopMid] < key) then
        chopMin = chopMid+1;
      else
        chopMax = chopMid;
      end
    end

    -- Check a match was found for the pet
    if ((chopMax == chopMin) and (petInfo[chopMin] == key)) then
      -- Update the pet's level
      petInfo[chopMax] = (petSpeciesId*4096)+(petNewLvl*16)+petQuality;
      table.sort(petInfo);

      -- Force the last pet info to be updated, if required
      if (lastPetInfo.speciesId == petSpeciesId) then
        lastPetInfo.speciesId = 0;
      end
    else
      C_Timer.After(1.0, NigglesPetsGetOwned);
    end
  end

  return;
end

-------------------------------------------------------------------------------
--                 L  O  C  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

-- NOTE: The following tables must be declared before the functions that use
--       them but initialised after the functions in them have been defined.

-- Special priority for mounts
-- NOTE: Entries MUST have 3 values representing run, fly and swim priorities
mountPriority =
{
  [SPELL_DRUID_CAT   ] = {1, 0, 0}, -- Cat Form
  [SPELL_DRUID_TRAVEL] = {1, 4, 4}, -- Travel Form
  [SPELL_SHAMAN_GHOST] = {1, 0, 0}, -- Ghost Wolf
  [SPELL_MANASABER   ] =            -- Arcanist's Manasaber
  {
    mountPriorityManasaber,
    mountPriorityManasaber,
    mountPriorityManasaber
  },
};

-- NOTE: The following table must be declared before the functions that use
--       them but initialised after the functions in them have been defined.
petJournalExtraButtons =
{
  -- Battle Pet Bandage
  {
    id         = 86143,
    type       = "item",
    typePrefix = "item:",
    setting    = "UiPetBandage",
    onEnter    = petBandageButtonOnEnter,
    onEvent    = petBandageButtonOnEvent,
    onHide     = petBandageButtonOnHide,
    onShow     = petBandageButtonOnShow,
  },
  -- Safari Hat
  {
    id         = 92738,
    type       = "toy",
    typePrefix = "",
    setting    = "UiPetSafariHat",
    onEnter    = petSafariHatButtonOnEnter,
    onEvent    = petSafariHatButtonOnEvent,
    onHide     = petSafariHatButtonOnHide,
    onShow     = petSafariHatButtonOnShow,
  },
};
