-------------------------------------------------------------------------------
--                L  O  C  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

local addonName, L = ...;

local ILVL_PATTERN        = "^"..string.gsub(_G.ITEM_LEVEL, "%%d", "(%%d+)");
local LOWLEVEL_MULTIPLIER = 0.8;
local MAX_PET_LEVEL       = 25;
local MAX_ACTIVE_PETS     = 3;

-- Names of the equipment slots
local equipSlotNames =
{
  [INVSLOT_HEAD    ] = "Head",
  [INVSLOT_NECK    ] = "Neck",
  [INVSLOT_SHOULDER] = "Shoulder",
  [INVSLOT_CHEST   ] = "Chest",
  [INVSLOT_WAIST   ] = "Waist",
  [INVSLOT_LEGS    ] = "Legs",
  [INVSLOT_FEET    ] = "Feet",
  [INVSLOT_WRIST   ] = "Wrist",
  [INVSLOT_HAND    ] = "Hands",
  [INVSLOT_FINGER1 ] = "Finger0",
  [INVSLOT_FINGER2 ] = "Finger1",
  [INVSLOT_TRINKET1] = "Trinket0",
  [INVSLOT_TRINKET2] = "Trinket1",
  [INVSLOT_BACK    ] = "Back",
  [INVSLOT_MAINHAND] = "MainHand",
  [INVSLOT_OFFHAND ] = "SecondaryHand",
};

-- Auction sub-categories for 2H items that aren't actually flagged as 2H
local equippedTwoHanded =
{
  ["Bows"     ] = true,
  ["Crossbows"] = true,
  ["Guns"]      = true,
};

-- Item IDs for all know lances
local tournamentLances =
{
  [46069] = true, -- Alliance Lance
  [46070] = true, -- Horde Lance
  [46106] = true, -- Argent Lance
  [52716] = true  -- Twilight Firelance
};

-- Info for Pet Battle items that grant an XP bonus
local petBattleItems =
{
  {toyId  =  92738, buffId = 158486}, -- Safari Hat
  {itemId =  98114, buffId = 142205}, -- Pet Treat
  {itemId =  98112, buffId = 142204}, -- Lesser Pet Treat
};

local paperDollInfo =
{
  ["Character"] =
  {
    indicators = {},
    stats =
    {
      average = 0,
      count   = 0,
      lowIlvl = 0,
      empty   = 0,
    }
  },
  ["Inspect"] =
  {
    indicators = {},
    stats =
    {
      average = 0,
      count   = 0,
      lowIlvl = 0,
      empty   = 0,
    },
    ilvlFrame = nil,
  },
};

local itemSlotRemap =
{
  -- Legion Artifacts
  [128293] = INVSLOT_MAINHAND, -- Death Knight, Frost
  [127830] = INVSLOT_MAINHAND, -- Demon Hunter, Havoc
  [128831] = INVSLOT_MAINHAND, -- Demon Hunter, Vengeance
  [128859] = INVSLOT_MAINHAND, -- Druid, Feral
  [128822] = INVSLOT_MAINHAND, -- Druid, Guardian
  [133959] = INVSLOT_MAINHAND, -- Mage, Fire
  [133948] = INVSLOT_MAINHAND, -- Monk, Windwalker
  [128866] = INVSLOT_OFFHAND,  -- Paladin, Protection
  [133958] = INVSLOT_MAINHAND, -- Priest, Shadow
  [128869] = INVSLOT_MAINHAND, -- Rogue, Assassination
  [134552] = INVSLOT_MAINHAND, -- Rogue, Outlaw
  [128479] = INVSLOT_MAINHAND, -- Rogue, Subtlety
  [128936] = INVSLOT_MAINHAND, -- Shaman, Elemental
  [128873] = INVSLOT_MAINHAND, -- Shaman, Enhancement
  [128934] = INVSLOT_MAINHAND, -- Shaman, Restoration
  [137246] = INVSLOT_OFFHAND,  -- Warlock, Demonology
  [134553] = INVSLOT_MAINHAND, -- Warrior, Fury
  [128288] = INVSLOT_OFFHAND,  -- Warrior, Protection
};

-- Pet Tamer IDs
local petTamerIds =
{
  [ 63194] = true, -- Steven Lisbane 
  [ 64330] = true, -- Julia Stevens  
  [ 65648] = true, -- Old MacDonald 
  [ 65651] = true, -- Lindsay
  [ 65655] = true, -- Eric Davidson 
  [ 65656] = true, -- Bill Buckler
  [ 66126] = true, -- Zunta
  [ 66135] = true, -- Dagra the Fierce 
  [ 66136] = true, -- Analynn 
  [ 66137] = true, -- Zonya the Sadist 
  [ 66352] = true, -- Traitor Gluk
  [ 66372] = true, -- Merda Stronghoof 
  [ 66412] = true, -- Elena Flutterfly 
  [ 66422] = true, -- Cassandra Kaboom 
  [ 66436] = true, -- Grazzle the Great 
  [ 66442] = true, -- Zoltan 
  [ 66452] = true, -- Kela Grimtotem 
  [ 66466] = true, -- Stone Cold Trixxy 
  [ 66478] = true, -- David Kosse
  [ 66512] = true, -- Deiza Plaguehorn 
  [ 66515] = true, -- Kortas Darkhammer 
  [ 66518] = true, -- Everessa
  [ 66520] = true, -- Durin Darkhammer 
  [ 66522] = true, -- Lydia Accoste 
  [ 66550] = true, -- Nicki Tinytech 
  [ 66551] = true, -- Ras'an
  [ 66552] = true, -- Narrok 
  [ 66553] = true, -- Morulu The Elder 
  [ 66557] = true, -- Bloodknight Antari 
  [ 66635] = true, -- Beegle Blastfuse 
  [ 66636] = true, -- Nearly Headless Jacob 
  [ 66638] = true, -- Okrut Dragonwaste 
  [ 66639] = true, -- Gutretch 
  [ 66675] = true, -- Major Payne 
  [ 66730] = true, -- Hyuna of the Shrines 
  [ 66733] = true, -- Mo'ruk 
  [ 66734] = true, -- Farmer Nishi 
  [ 66738] = true, -- Courageous Yon 
  [ 66739] = true, -- Wastewalker Shu 
  [ 66741] = true, -- Aki the Chosen 
  [ 66815] = true, -- Bordin Steadyfist 
  [ 66819] = true, -- Brok 
  [ 66822] = true, -- Goz Banefury 
  [ 66824] = true, -- Obalis 
  [ 66918] = true, -- Seeker Zusshi 
  [ 67370] = true, -- Jeremy Feasel 
  [ 68462] = true, -- Flowing Pandaren Spirit 
  [ 68463] = true, -- Burning Pandaren Spirit 
  [ 68464] = true, -- Whispering Pandaren Spirit 
  [ 68465] = true, -- Thundering Pandaren Spirit 
  [ 71924] = true, -- Wrathion 
  [ 71926] = true, -- Lorewalker Cho
  [ 71927] = true, -- Chen Stormstout
  [ 71929] = true, -- Sully "The Pickle" McLeary
  [ 71930] = true, -- Shademaster Kiryn
  [ 71931] = true, -- Taran Zhu
  [ 71932] = true, -- Wise Mari
  [ 71933] = true, -- Blingtron 4000
  [ 71934] = true, -- Dr. Ion Goldbloom 
  [ 73626] = true, -- Little Tommy Newcomer
  [ 79179] = true, -- Squirt
  [ 83837] = true, -- Cymre Brightblade 
  [ 85519] = true, -- Christoph VonFeasel 
  [ 87110] = true, -- Tarr the Terrible 
  [ 87122] = true, -- Gargra 
  [ 87123] = true, -- Vesharr 
  [ 87124] = true, -- Ashlei 
  [ 87125] = true, -- Taralune 
  [ 90675] = true, -- Erris the Collector
  [ 91014] = true, -- 
  [ 91015] = true, -- 
  [ 91016] = true, -- 
  [ 91017] = true, -- 
  [ 97804] = true, -- Tiffany Nelson
  [ 98270] = true, -- Robert Craig
  [ 99035] = true, -- Durian Strongfruit 
  [ 99077] = true, -- Bredda Tenderhide 
  [ 99150] = true, -- Grixis Tinypop
  [ 99182] = true, -- Sir Galveston
  [ 99210] = true, -- Bodhi Sunwayver 
  [ 99878] = true, -- Ominitron Defense System
  [104553] = true, -- Odrogg 
  [104970] = true, -- Xorvasc 
  [105093] = true, -- Fragment of Fire
  [105250] = true, -- Aulier
  [105323] = true, -- Ancient Catacomb Eggs
  [105386] = true, -- Rydyr 
  [105387] = true, -- Andurs 
  [105455] = true, -- Trapper Jarrun
  [105512] = true, -- Envoy of the Hunt 
  [105674] = true, -- Varenne 
  [105779] = true, -- Felsoul Seer
  [106552] = true, -- Nightwatcher Merayl
  [107489] = true, -- Amalia
  [115286] = true, -- Crysa
  [115307] = true, -- Algalon the Observer
  [117934] = true, -- Sissix
  [117950] = true, -- Madam Viciosa
  [117951] = true, -- Nameless Mystic
  [124617] = true, -- Environeer Bert
};

local missingPetXPBuffs = nil;

local lockedItemInfo =
{
  loc  = ItemLocation:CreateEmpty()
};

local gameTooltipInfo =
{
  textLeft  = {},
  lastItem  =
  {
    link      = 0,
    isUnknown = false,
  },
};

-------------------------------------------------------------------------------
--             L  O  C  A  L     D  E  F  I  N  I  T  I  O  N  S
-------------------------------------------------------------------------------

local getItemLinkLevel;
local petBattleEquipmentGetMissing;

-------------------------------------------------------------------------------
--                L  O  C  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to add a message to chat frames
--
local function addChatMessage(msg, chatType)
  -- Local Variables
  local isAdded = false;
  local colorInfo = ChatTypeInfo[chatType];

  -- Add the notification to any appropriate chat frames
  for _, current in pairs(CHAT_FRAMES) do
    chatFrame = _G[current];
    chatId = chatFrame:GetID();
    if ((chatId >= 1) and 
        (chatId <= NUM_CHAT_WINDOWS) and 
        (tContains({GetChatWindowMessages(chatId)}, "SYSTEM"))) then
      -- Add the notification the chat frame
      chatFrame:AddMessage(msg, colorInfo.r, colorInfo.g, colorInfo.b,
        colorInfo.id);
      isAdded = true;
    end
  end
  
  -- If the message hasn't been added to any chat frame...
  if (not isAdded) then
    -- ...add it to the default chat frame
    DEFAULT_CHAT_FRAME:AddMessage(msg, colorInfo.r, colorInfo.g,
      colorInfo.b, colorInfo.id);
  end

  -- Play a sound to alert the user
  PlaySoundFile("Sound\\Interface\\UI_Garrison_Mission_Complete_MissionFail_Stinger.ogg");

  return;
end

--
-- Function to get the item level for an item from a tooltip.
-- NOTE: This function is required because 'GetItemInfo()' and even
--       'GetDetailedItemLevelInfo()' do not return reliable item levels,
--        especially for items with modifiers such as 'Mythic', 'Titanforged'
--        or 'Timewalking'.
--
getItemLinkLevel = function(itemLink)
  -- Local Variables
  local _;
  local itemIlvl = nil;
  local itemName
  local itemQuality;
  local textLeft = NigglesTooltip.textLeft;
  local numLines

  -- Set the tooltip and scan its lines for the item level
  numLines = NigglesTooltip:SetContents("SetHyperlink", itemLink);
  for lineIdx = 2, math.min(numLines, 5) do
    _, _, itemIlvl = string.find(textLeft[lineIdx]:GetText(), ILVL_PATTERN);
    if (itemIlvl ~= nil) then
      itemIlvl = tonumber(itemIlvl);
      break;
    end
  end

  return itemIlvl;
end

--
-- Function called when the Gossip panel is hidden. If a warning about missing
-- pet XP buffs is displayed, the current time is saved. The time is used to
-- determine if a warning should be displayed when a pet battle starts.
--
local function gossipOnHide()
  -- Check if the warning about missing pet XP buffs exists
  if (missingPetXPBuffs ~= nil) then
    -- Check if the warning about missing pet XP buffs is shown
    if (missingPetXPBuffs:IsShown()) then
      missingPetXPBuffs.time = GetTime();
    end

    -- Hide the warning about missing pet XP buffs
    missingPetXPBuffs:Hide();
  end

  return;
end

--
-- Function called when the options are changed in the Gossip frame
--
local function gossipOptionsOnUpdate(...)
  -- Local Variables
  local buttonIdx = NUMGOSSIPBUTTONS;
  local child;
  local frame = missingPetXPBuffs;
  local missingEquipment;

  -- Check if the current NPC is a pet tamer
  if ((NigglesDBC.settings.WarningPetBattle) and
      (L.npcIsPetTamer())) then
    missingEquipment = petBattleEquipmentGetMissing(true);
    if (missingEquipment ~= "") then
      -- Create the missing equipment frame, if required
      if (frame == nil) then
        frame = CreateFrame("FRAME", nil, GossipGreetingScrollChildFrame);
        frame:SetSize(1,1);
        frame:SetPoint("LEFT", 8, 0);
        frame:SetPoint("RIGHT", GossipGreetingText, "RIGHT");
        missingPetXPBuffs = frame;
        child = frame:CreateTexture(nil, "ARTWORK", nil);
        child:SetTexture("Interface\\DialogFrame\\UI-Dialog-Icon-AlertNew");
        child:SetPoint("TOPLEFT");
        child:SetSize(40, 40);
        frame.icon = child;
        child = frame:CreateFontString(nil, "ARTWORK");
        child:SetFontObject(GossipTitleButton1:GetNormalFontObject());
        child:SetJustifyH("LEFT");
        child:SetPoint("TOPRIGHT", 0, -8);
        child:SetPoint("LEFT", frame.icon, "RIGHT", 8, 0);
        child:SetText(L["PetBattleMissingBuffs"]);
        frame.label = child;
        child = frame:CreateFontString(nil, "ARTWORK");
        child:SetFontObject(GossipTitleButton1:GetNormalFontObject());
        child:SetJustifyH("LEFT");
        child:SetPoint("TOPLEFT", frame.label, "BOTTOMLEFT", 12, -8);
        child:SetPoint("RIGHT");
        frame.list = child;
      end

      -- Find the last visible gossip title
      while ((buttonIdx > 0) and
             (not _G["GossipTitleButton"..buttonIdx]:IsShown())) do
        buttonIdx = buttonIdx-1;
      end

      -- Show the missing equipment frame
      frame.list:SetText(missingEquipment);
      frame:SetPoint("TOP", _G["GossipTitleButton"..buttonIdx], "BOTTOM",
        0, -_G["GossipTitleButton"..buttonIdx]:GetHeight());
      frame:Show();

      -- Save the GUID of the NPC
      missingPetXPBuffs.npcGuid = UnitGUID("npc");
    end
  elseif (frame ~= nil) then
    frame:Hide();
  end

  return;
end

--
-- Function to update the item level of the inspected unit
--
local function inspectedUnitUpdateIlvl()
  -- Local Variables
  local ilvlFrame = paperDollInfo["Inspect"].ilvlFrame;
  local indicator;
  
  -- Check if the inspected character's item level should be displayed
  if ((NigglesDBC.settings.UiInspectedItemLevel) and
      (InspectFrame.unit ~= nil) and
      (UnitLevel(InspectFrame.unit) >=
        MIN_PLAYER_LEVEL_FOR_ITEM_LEVEL_DISPLAY)) then
    -- Scan the inspected character's equipment and calculate the item level
    L.equipmentScan("Inspect");

    -- Update the item level displayed
    ilvlFrame.text:SetText(
      STAT_AVERAGE_ITEM_LEVEL..": "..HIGHLIGHT_FONT_COLOR_CODE..
      math.ceil(paperDollInfo["Inspect"].stats.average));
    ilvlFrame:Show();
  elseif (ilvlFrame ~= nil) then
    -- Hide the item level displayed
    ilvlFrame:Hide();
   
    -- Hide all the low ilvl indicators
    for _, indicator in pairs(paperDollInfo["Inspect"].indicators) do
      indicator:Hide();
    end
  end

  return;
end

--
-- Function called when the inspected unit's inventory changes
--
local function inspectedUnitOnEvent(self, event, unit)
  -- Check if the event should be processed
  if ((NigglesDBC.settings.UiInspectedItemLevel) and
      (InspectFrame:IsVisible())) then
    -- Process the event
    if (((event == "UNIT_INVENTORY_CHANGED") and
         (unit == InspectFrame.unit)) or
        (event == "GET_ITEM_INFO_RECEIVED")) then
      inspectedUnitUpdateIlvl();
    end
  end

  return;
end

--
-- Function to check if any appearance an item has is known. If the item
-- is soul-bound or doesn't have an appearance then it is considered known.
--
local function isAppearanceUnknown(itemId)
  -- Local Variables
  local _;
  local appearanceId;
  local isUnknown = IsDressableItem(itemId);
  local itemSubType;
  local itemSubType;
  local itemType;
  local sourceId;
  local sourceIds;
  local sourceInfo;
  local sourceSubType
  local sourceType

  -- Check if the item has an appearance
  if (isUnknown) then
    appearanceId = C_TransmogCollection.GetItemInfo(itemId);
    if (appearanceId == null) then
      isUnknown = false;
    end
  end

  -- Check for any known source IDs for the item's appearance
  if (isUnknown) then
    -- Get the item's type and sub-type
    _, itemType, itemSubType = GetItemInfoInstant(itemId);

    -- Get all the source IDs
    sourceIds = C_TransmogCollection.GetAllAppearanceSources(appearanceId);
    if (sourceIds ~= nil) then
      -- Process each source
      for _, sourceId in ipairs(sourceIds) do
        -- Check if the source is known
        sourceInfo = C_TransmogCollection.GetSourceInfo(sourceId);
        if ((sourceInfo ~= nil) and (sourceInfo.isCollected)) then
          -- Check the source type is the same as the item type
          -- NOTE: Added for edge cases of the same appearance from a different
          --       type of armour
          _, sourceType, sourceSubType = GetItemInfoInstant(sourceInfo.itemID);
          if ((sourceType    == itemType   ) and
              (sourceSubType == itemSubType)) then
            isUnknown = false;
            break;
          end
        end
      end
    end
  end

  return isUnknown;
end

--
-- Function called when an item is being sold
--
local function itemOnSold(container, slot, target)
  -- Local Variables
  local _;
  local warning = nil;

  -- Check if the item is being sold
  if ((MerchantFrame:IsShown()) and
      (MerchantFrame.selectedTab == 1) and
      (lockedItemInfo.id ~= nil)) then
    -- Check if the locked item is part of an equipment set
    if ((NigglesDBC.settings.WarningEquipSetItem) and
        (lockedItemInfo.inEquipSet)) then
      warning = L["EquipSetItemSold"];
    end

    -- Check if the locked item's appearance is uncollected
    if ((NigglesDBC.settings.WarningUncollectedSold) and
        (lockedItemInfo.id ~= nil) and
        (lockedItemInfo.appearanceUnknown)) then
      warning = L["UncollectedAppearanceSold"];
    end
   
    -- Display a warning, if required
    if (warning ~= nil) then
      addChatMessage(warning, "SYSTEM");
    end
  end

  return;
end

--
-- Function to process a button click in a warning popup for missing pet battle
-- XP buffs
--
local function petBattleBuffsOnClick(popup, id)
  -- Check if the 'Forfeit' button was clicked
  if (id == 2) then
    C_PetBattles.ForfeitGame();
  end

  -- Hide the popup
  StaticPopupSpecial_Hide(popup);

  return;
end

--
-- Function to check for Pet Battle equipment that isn't equipped, if there
-- are any pets that aren't maximum level.
--
-- NOTE: This function will need enhancement if Blizzard ever add items for
--       ring, trinket, main or off-hand slots.
--
petBattleEquipmentGetMissing = function(simple)
  -- Local Variables
  local _;
  local auras = L.playerAuras.get();
  local forfeitPenalty;
  local itemLink;
  local itemName;
  local itemTexture;
  local nonMaxLevel = false;
  local petGUID;
  local petLevel;
  local string = "";
  local idx;

  -- Work out if there are any pets that aren't max level
  for idx = 1, MAX_ACTIVE_PETS do
    petGUID = C_PetJournal.GetPetLoadOutInfo(idx);
    if (petGUID ~= nil) then
      _, _, petLevel = C_PetJournal.GetPetInfoByPetID(petGUID);
      if (petLevel < MAX_PET_LEVEL) then
        nonMaxLevel = true;
        break;
      end
    end
  end

  -- Check if there are any pets that aren't max level
  if (nonMaxLevel) then
    -- Get the localized names for the pet battle spells, if required
    if (petBattleItems[1].buffName == nil) then
      for _, itemInfo in ipairs(petBattleItems) do
        itemInfo.buffName = GetSpellInfo(itemInfo.buffId);
      end
    end

    -- Check if there are any XP buffs missing that are available
    for itemIdx, itemInfo in ipairs(petBattleItems) do
      if (auras[itemInfo.buffName] ~= true) then
        -- Get info for the missing XP buff
        itemLink    = nil;
        itemTexture = nil;
        if ((itemInfo.itemId ~= nil) and
            (GetItemCount(itemInfo.itemId, false) > 0)) then
          itemName, itemLink, _, _, _, _, _, _, _, itemTexture =
            GetItemInfo(itemInfo.itemId);
        elseif ((itemInfo.toyId ~= nil) and
                (PlayerHasToy(itemInfo.toyId))) then
          _, itemName, itemTexture = C_ToyBox.GetToyInfo(itemInfo.toyId);
          itemLink = C_ToyBox.GetToyLink(itemInfo.toyId);
        end

        -- Check the info for the missing XP buff could be obtained
        if ((itemTexture ~= nil) and (itemLink ~= nil)) then
          if (simple) then
            string = string..itemName.."\n";
          else
            -- NOTE: The '|r' is done after the '\n' to overcome a Blizzard bug
            string = string.."\124T"..itemTexture..":0:0:0:-8\124t "..
              itemLink.."\n|r";
          end
        end
      end
    end
  end

  return string;
end

--
-- Function called when the GameTooltip is hidden. It clears info about the
-- last item displayed, so that if an item is equipped, its collected status
-- will be updated.
--
local function uncollectedOnHide()
  -- Clear the info for the last item
  gameTooltipInfo.lastItem.link      = "";
  gameTooltipInfo.lastItem.isUnknown = false;

  return;
end

--
-- Function called when GameTooltip is updated to display the info for an item.
-- It checks if the appearance of the item is uncollected and adds a line to
-- GameTooltip if so.
--
local function uncollectedOnTooltipSetItem(tooltip)
  -- Local Variables
  local _;
  local itemLink;
  local itemType;
  local lastItem = gameTooltipInfo.lastItem;
  local lineIdx;
  local lineText;
  local numLines = tooltip:NumLines();
  local textLeft = gameTooltipInfo.textLeft;

  -- Check the feature is enabled and the game isn't already implementing it
  if ((NigglesDBC.settings.UiAllUncollected) and
      (not C_TransmogCollection.GetShowMissingSourceInItemTooltips())) then
    -- Check if the item's info has already been obtained
    _, itemLink = tooltip:GetItem();
    if (itemLink ~= lastItem.link) then
      -- Initialise the info for the last item
      lastItem.link      = itemLink;
      lastItem.isUnknown = false;

      -- Check the item is dressable
      itemType = select(6, GetItemInfoInstant(itemId));
      if (((itemType == 2) or   -- Weapons
           (itemType == 4)) and -- Armor
          (IsDressableItem(itemLink))) then
        -- Get handles to any new text lines in GameTooltip
        lineIdx = #textLeft;
        while (lineIdx <= numLines) do
          textLeft[lineIdx] = _G["GameTooltipTextLeft"..lineIdx];
          lineIdx = lineIdx+1;
        end

        -- Check the tooltip for 'Soulbound' or an existing warning 
        lineIdx = 1;
        while (lineIdx <= numLines) do
          lineText = textLeft[lineIdx]:GetText();
          if ((lineText == ITEM_SOULBOUND) or
              (lineText == TRANSMOGRIFY_TOOLTIP_APPEARANCE_UNKNOWN)) then
            break;
          end
          lineIdx = lineIdx+1;
        end

        -- Check if the appearance is unknown
        if (lineIdx > numLines) then
          lastItem.isUnknown = isAppearanceUnknown(itemLink);
        end
      end
    end
   
    -- Add a line to the tooltip if the item's appearance is unknown
    if (lastItem.isUnknown) then
      tooltip:AddLine(TRANSMOGRIFY_TOOLTIP_APPEARANCE_UNKNOWN,
        0.533, 0.666, 1, 1);
    end
  end

  return;
end

-------------------------------------------------------------------------------
--                 A  D  D  O  N     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to check for Pet Battle items aren't equipped at start of a battle.
--
function L.equipmentCheckPetBattleXPBuffs()
  -- Local Variables
  local _;
  local forfeitPenalty;
  local missingEquipment;

  -- Check if there are any missing buffs
  if ((missingPetXPBuffs == nil) or
      (missingPetXPBuffs.time == nil) or
      (GetTime()-missingPetXPBuffs.time > 2)) then
    missingEquipment = petBattleEquipmentGetMissing();
    if (missingEquipment ~= "") then
      -- Display a warning about the missing buffs
      forfeitPenalty = C_PetBattles.GetForfeitPenalty();
      NigglesWarningPopupInit(L["PetBattleMissingBuffs"],
        (forfeitPenalty > 0 and
          format(L["PetBattleHealthLoss"], forfeitPenalty) or nil),
        CONTINUE, PET_BATTLE_FORFEIT, petBattleBuffsOnClick);
      NigglesWarningPopup.list:SetText(missingEquipment);
      NigglesWarningPopup.list:Show();
      StaticPopupSpecial_Show(NigglesWarningPopup);
    end
  end

  return;
end

--
-- Function to dump debug info for the equipment features
--
function L.equipmentDumpDebug()
  -- Local Variables
  local info = ChatTypeInfo["SYSTEM"];
  local key;
  local value;

  -- Dump the equipment stats for the player
  DEFAULT_CHAT_FRAME:AddMessage("    Ngl-equipment:",
    info.r, info.g, info.b);
  DEFAULT_CHAT_FRAME:AddMessage("        Player Equipment Stats:",
    info.r, info.g, info.b);
  for key, value in pairs(paperDollInfo["Character"].stats) do
    DEFAULT_CHAT_FRAME:AddMessage("            "..key..": "..(value or ""),
    info.r, info.g, info.b);
  end

  return;
end

--
-- Function to get the number of empty slots
--
function L.equipmentGetEmptySlots()
  return paperDollInfo["Character"].stats.empty;
end

--
-- Function to get the number of low level items equipped
--
function L.equipmentGetLowIlvl()
  return paperDollInfo["Character"].stats.lowIlvl;
end

--
-- Function called when the Blizzard Inspect UI addon.
--
function L.equipmentInspectUIOnLoad()
  -- Local Variables
  local frame;
  local child;

  -- Hook the function that updates the buttons in the Inspect panel
  hooksecurefunc("InspectPaperDollFrame_UpdateButtons",
    inspectedUnitUpdateIlvl);

  -- Create a frame to display the item level of the inspected unit
  frame = CreateFrame("FRAME", nil, InspectPaperDollItemsFrame);
  frame:SetSize(100, 10);
  frame:SetPoint("BOTTOMRIGHT", -14, 14);
  paperDollInfo["Inspect"].ilvlFrame = frame;
  child = frame:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall");
  child:SetPoint("TOPRIGHT");
  child:SetJustifyH("RIGHT");
  frame.text = child;
  frame:SetScript("OnEvent", inspectedUnitOnEvent);
  frame:RegisterEvent("GET_ITEM_INFO_RECEIVED");
  frame:RegisterEvent("UNIT_INVENTORY_CHANGED");

  return;
end

--
-- Function to check if a lance is equipped
--
function L.equipmentIsLance(itemId)
  return (tournamentLances[itemId] ~= nil);
end

--
-- Function to save info for an item in a container when it is locked. The
-- info is used when items are sold.
--
function L.equipmentSaveItemInfo(containerId, slotId)
  -- Clear the ID for the locked item
  lockedItemInfo.id = nil;

  -- Check the locked item is in a container
  if ((containerId ~= nil) and (slotId ~= nil)) then
    -- Get info for the item
    _, _, _, _, _, _, lockedItemInfo.link, _, _, lockedItemInfo.id =
      GetContainerItemInfo(containerId, slotId);
    if (lockedItemInfo.id ~= nil) then
      -- Extract the item's name from the locked item's link
      _, _, lockedItemInfo.name = string.find((lockedItemInfo.link or ""),
        "|h%[(.*)%]|h|r$");

      -- Check if the locked item is in an equipment set
      lockedItemInfo.inEquipSet =
        GetContainerItemEquipmentSetInfo(containerId, slotId);

      -- Check if the locked item's appearance is unknown
      lockedItemInfo.appearanceUnknown = false;
      if (IsDressableItem(lockedItemInfo.id)) then
        lockedItemInfo.loc:SetBagAndSlot(containerId, slotId);
        if (not C_Item.IsBound(lockedItemInfo.loc)) then
          lockedItemInfo.appearanceUnknown =
            isAppearanceUnknown(lockedItemInfo.id);
        end
      end
    end
  end

  return;
end

--
-- Function to scan all equipped items. It works out an average ilvl, as well
-- as counting the number of empty slots.
--
function L.equipmentScan(target)
  -- Local Variables
  local _;
  local equippedIlvl = {};
  local indicators = paperDollInfo[target].indicators;
  local infoAvailable = true;
  local itemLink;
  local itemQuality;
  local itemSubCat;
  local itemType;
  local itemLvl;
  local lowIlvl;
  local stats = paperDollInfo[target].stats;
  local totalIlvl = 0;
  local totalIlvl = 0;
  local unit;

  -- Initialise some variables
  unit = ((target == "Inspect") and InspectFrame.unit or "player");

  -- Reset the stats for the equipped items
  stats.average = 0;
  stats.count   = 0;
  stats.lowIlvl = 0;
  stats.empty   = 0;

  -- Scan all equipment slots
  for slotId, slotName in pairs(equipSlotNames) do
    -- Initialise some variables
    equippedIlvl[slotId] = 0;

    -- Check if the slot is contains an item
    itemLink = GetInventoryItemLink(unit, slotId);
    if (itemLink ~= nil) then
      -- Check if the item's level should be taken from another slot
      -- NOTE: This is to deal with a bug in the Blizzard API that means
      --       Legion artifact weapons that use both the main and off-hand
      --       slots will always give an item level of 750 for one of them.
      _, _, itemId = string.find(itemLink, "|Hitem:(%d+):");
      if (itemId ~= nil) then
        itemId = tonumber(itemId);
        if (itemSlotRemap[itemId] ~= nil) then
          itemLink = GetInventoryItemLink(unit, itemSlotRemap[itemId]);
        end
      end

      -- Get the item's info
      _, _, itemQuality = GetItemInfo(itemLink);
      itemLvl           = getItemLinkLevel(itemLink);
      if ((itemQuality ~= nil) and (itemLvl ~= nil)) then
        -- Update the info for calculating the average item level
        equippedIlvl[slotId] = itemLvl;
        totalIlvl            = totalIlvl+itemLvl;
        stats.count          = stats.count+1;
      else
        infoAvailable = false;
        break;
      end
    end
  end

  -- Check item info could be obtained for all items
  if (infoAvailable) then
    -- Check if the info needs adjusting to account for 2H weapons
    -- NOTE: The Blizzard API average item level calculation treats 2H weapons
    --       as if there was a 2nd weapon in the off-hand of same item level.
    if ((equippedIlvl[INVSLOT_OFFHAND] == 0) and
        (equippedIlvl[INVSLOT_MAINHAND] > 0)) then
      -- Get the main hand's type and sub category
      _, _, _, _, _, _, itemSubCat, _, itemType =
        GetItemInfo(GetInventoryItemLink(unit, INVSLOT_MAINHAND));
      if ((itemType == "INVTYPE_2HWEAPON") or
          (equippedTwoHanded[itemSubCat] ~= nil)) then
        -- Set the off-hand to the same item level as the main hand
        equippedIlvl[INVSLOT_OFFHAND] = equippedIlvl[INVSLOT_MAINHAND];

        -- Update the info for calculating the average item level
        totalIlvl   = totalIlvl+equippedIlvl[INVSLOT_MAINHAND];
        stats.count = stats.count+1;
      end
    end

    -- Work out the average item level
    if (stats.count > 0) then
      stats.average = totalIlvl/stats.count;
    else
      stats.average = 0;
    end

    -- Work out how many low level items are equipped
    lowIlvl = stats.average*LOWLEVEL_MULTIPLIER;
    for slotId, slotIlvl in pairs(equippedIlvl) do
      if (slotIlvl == 0) then
        stats.empty = stats.empty+bit.lshift(1, slotId);
      elseif (slotIlvl < lowIlvl) then
        stats.lowIlvl = stats.lowIlvl+bit.lshift(1, slotId);
      end
    end
  else
    -- Reset the stats for the equipped items
    stats.count   = 0;
    stats.lowIlvl = 0;
    stats.empty   = 0;

    -- Scan the equipment again in a short while
    C_Timer.After(0.5,
      function()
        L.equipmentScan(target);
      end);
  end

  -- Update the low level item indicators
  if ((NigglesDBC.settings.UiLowLevelItems) and
      (stats.lowIlvl > 0)) then
    lowIlvl = stats.average*LOWLEVEL_MULTIPLIER;
    for slotId, slotIlvl in pairs(equippedIlvl) do
      if ((slotIlvl ~= 0) and
          (slotIlvl < lowIlvl)) then
        -- Create the equipment slot's low Ilvl indicator, if required
        -- NOTE: A font string can't be created as a child of the equipment
        --       slot because it can't be placed above the cooldown.
        if (indicators[slotId] == nil) then
          indicators[slotId] = CreateFrame("Frame", nil,
            _G[target..equipSlotNames[slotId].."Slot"],
            "NigglesLowIlvlTemplate");
        end
        indicators[slotId]:Show();
      elseif (indicators[slotId] ~= nil) then
        indicators[slotId]:Hide();
      end
    end
  else
    -- Hide all the low ilvl indicators
    for _, indicator in pairs(indicators) do
      indicator:Hide();
    end
  end

  return;
end

--
-- Function to check if the current NPC is a pet tamer
--
function L.npcIsPetTamer()
  -- Local Variables
  local npcId = tonumber(string.match((UnitGUID("npc") or ""),
                  "[^-]+-[^-]+-[^-]+-[^-]+-[^-]+-([^-]+)-"));

  return ((npcId ~= nil) and (petTamerIds[npcId] == true));
end

-------------------------------------------------------------------------------
--               G  L  O  B  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to initialise the Niggles Equipment functions
--
function L.equipmentInit()
  -- Local Variables
  local idx;

  -- Hook the function used to sell items
  hooksecurefunc("UseContainerItem", itemOnSold);

  -- Hook the function used to update the gossip frame
  hooksecurefunc("GossipFrameOptionsUpdate", gossipOptionsOnUpdate);
  for idx = 1, NUMGOSSIPBUTTONS do
    _G["GossipTitleButton"..idx]:Hide();
  end
  GossipFrame:HookScript("OnHide", gossipOnHide);

  -- Hook GameTooltip for uncollected appearances
  GameTooltip:HookScript("OnTooltipSetItem", uncollectedOnTooltipSetItem);
  GameTooltip:HookScript("OnHide",           uncollectedOnHide          );

  return;
end

--
--
-- Function to enable/disable the equipment features
--
function L.equipmentEnable()
  -- (Un)register for events
  if ((NigglesDBC.settings.UiLowLevelItems) or
      (NigglesDBC.settings.WarningLowLevelItems)) then
    Niggles:RegisterEvent("PLAYER_EQUIPMENT_CHANGED");
  else
    Niggles:UnregisterEvent("PLAYER_EQUIPMENT_CHANGED");
  end

  -- Scan the equipped items
  L.equipmentScan("Character");
  if (InspectFrame ~= nil) then
    inspectedUnitUpdateIlvl();
  end

  return;
end
