-------------------------------------------------------------------------------
--                L  O  C  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

local addonName, L = ...;

-------------------------------------------------------------------------------
--                A  D  D  O  N     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

L.flagHolidayBrewfest        = 0x00000001;
L.flagHolidayChildrensWeek   = 0x00000002;
L.flagHolidayDarkmoonFaire   = 0x00000004;
L.flagHolidayDayOfTheDead    = 0x00000008;
L.flagHolidayHallowsEnd      = 0x00000010;
L.flagHolidayHarvestFestival = 0x00000020;
L.flagHolidayLoveInTheAir    = 0x00000040;
L.flagHolidayLunarFestival   = 0x00000080;
L.flagHolidayMidsummer       = 0x00000100;
L.flagHolidayNobleGarden     = 0x00000200;
L.flagHolidayPiratesDay      = 0x00000400;
L.flagHolidayWinterVeil      = 0x00000800;

-------------------------------------------------------------------------------
--                L  O  C  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

local holidayInfo =
{
  [235442] = L.flagHolidayBrewfest,
  [235445] = L.flagHolidayChildrensWeek,
  [235448] = L.flagHolidayDarkmoonFaire,
  [307365] = L.flagHolidayDayOfTheDead,
  [235462] = L.flagHolidayHallowsEnd,
  [235465] = L.flagHolidayHarvestFestival,
  [235468] = L.flagHolidayLoveInTheAir,
  [235471] = L.flagHolidayLunarFestival,
  [235474] = L.flagHolidayMidsummer,
  [235477] = L.flagHolidayNobleGarden,
  [235481] = L.flagHolidayPiratesDay,
  [235485] = L.flagHolidayWinterVeil,
};

local activeInfo =
{
  expires = 0,
  mask    = 0,
};

-------------------------------------------------------------------------------
--                 L  O  C  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to turn a date into a single numerical value that can be compared.
-- The components of the date are bit-shifted, using multiplication, such that
-- each component doesn't overlap any other element. The year is reduced by
-- 2000 to keep the overall value within 32 bits. The resultant value is safe
-- for simple comparisons but can't be used for any date arithmetic.
--
local function dateToTocks(date)
  return ((date.year-2000)*0x000100000)+
          (date.month     *0x000010000)+
          (date.monthDay  *0x000000800)+
          (date.hour      *0x000000040)+
           date.minute;
end

-------------------------------------------------------------------------------
--                 A  D  D  O  N     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to get a bit mask of the active holidays
--
function L.holidaysGetActive()
  -- Local Variables
  local activeFlags = 0;
  local currentDate = C_Calendar.GetDate();
  local currentTocks;
  local endTocks;
  local expires;
  local holidayInfo;
  local idx = 0;
  local startTocks;

  -- Convert the current date into tocks
  currentTocks = dateToTocks(currentDate);

  -- Check if the bit mask should be updated
  if (activeInfo.expires <= currentTocks) then
    -- Work out the default expires tocks
    currentDate.hour   = 23;
    currentDate.minute = 59;
    expires = dateToTocks(currentDate);

    -- Create a bit mask of the active holidays
    holidayInfo = C_Calendar.GetHolidayInfo(0, currentDate.monthDay, idx);
    while (holidayInfo ~= nil) do
      -- Check if the holiday is known
      if (holidayInfo[holidayInfo.texture] ~= nil) then
        -- Check the holiday is currently active
        startTocks = dateToTocks(holidayInfo.startTime);
        endTocks   = dateToTocks(holidayInfo.endTime);
        if ((currentTocks >= startTocks) and
            (currentTocks <  endTocks  )) then
          activeFlags = bit.bor(activeFlags, holidayInfo[holidayInfo.texture]);
        end

        -- Update the expires tocks
        if (startTocks > currentTocks) then
          expires = math.min(expires, startTocks);
        elseif (endTocks > currentTocks) then
          expires = math.min(expires, endTocks);
        end
      end

      -- Move on to the next holiday
      idx = idx+1;
      holidayInfo = C_Calendar.GetHolidayInfo(0, currentDate.monthDay, idx);
    end

    -- Update the active holiday info
    activeInfo.expires = expires;
    activeInfo.mask    = activeFlags;
  end

  return activeInfo.mask;
end
