-------------------------------------------------------------------------------
--                L  O  C  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

local addonName, L = ...;

local autoLootSlots = {};

local stackSplitMax;
local stackSplitMaxPurchasable;

-------------------------------------------------------------------------------
--             L  O  C  A  L     D  E  F  I  N  I  T  I  O  N  S
-------------------------------------------------------------------------------

local stackSplitMaxOnClick;

-------------------------------------------------------------------------------
--                L  O  C  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to auto-loot items with a short delay between each
--
local function autoLootNext()
  -- Local Variables
  local _;
  local slotIdx;
  local isLocked;
  local lootIdx;

  -- Search for an item that can be looted
  while ((next(autoLootSlots)) and (lootIdx == nil)) do
    slotIdx = autoLootSlots[#autoLootSlots];
    _, _, _, _, _, isLocked = GetLootSlotInfo(slotIdx);
    if (isLocked == false) then
      lootIdx = slotIdx;
    end
    autoLootSlots[#autoLootSlots] = nil;
  end

  -- If a suitable item was found, loot it
  if (lootIdx ~= nil) then
    LootSlot(lootIdx);
  end

  -- Loot another item after a short delay, if required
  if (next(autoLootSlots)) then
    C_Timer.After(LOOTFRAME_AUTOLOOT_DELAY, autoLootNext);
  end

  return;
end

--
-- Function to delete an item in a bag or bank slot with a modified click.
-- The deletion is confirmed by calling the appropriate Blizzard function.
--
local function itemDelete(bagId, slotId, mouseButton)
  -- Local Variables
  local _;

  -- Check the option is enabled and the item was Alt-Right clicked
  if ((NigglesDBC.settings.UiItemDelete == true) and
      (mouseButton == "RightButton") and
      (IsAltKeyDown())) then
    -- Get info for the item clicked on
    local itemId = GetContainerItemID(bagId, slotId);

    -- Check the item's ID could be obtained
    if (itemId) then
      -- Pick up the item clicked on
      PickupContainerItem(bagId, slotId);

      -- Get the quality of the item clicked on
      local itemName, _, itemQuality = GetItemInfo(itemId);

      -- Trigger the standard delete item action
      UIParent_OnEvent(self, "DELETE_ITEM_CONFIRM", itemName, itemQuality);
    end
  end

  return;
end

--
-- Function to delete an item in a bag with a modified click.
--
local function bagItemClicked(self, mouseButton)
  -- Delete the item
  itemDelete(self:GetParent():GetID(), self:GetID(), mouseButton);

  return;
end

--
-- Function to delete an item in a bank slot with a modified click.
--
local function bankItemClicked(self, mouseButton)
  -- Delete the item
  itemDelete(BANK_CONTAINER, self:GetID(), mouseButton);

  return;
end

--
-- Function to zoom the minimap all the way in when the button is
-- right-clicked.
--
local function minimapZoomIn(self, button)
  -- Check the option is enabled
  if (NigglesDBC.settings.UiMiniMapZoom == true) then
    -- Check the right button was clicked and the 'Zoom In' button is enabled
    if ((button == "RightButton") and
        (MinimapZoomIn:IsEnabled())) then
      MinimapZoomIn:Disable();
      MinimapZoomOut:Enable();
      PlaySound(SOUNDKIT.IG_MINIMAP_ZOOM_IN);
      Minimap:SetZoom(Minimap:GetZoomLevels()-1);
    end
  end

  return;
end

--
-- Function to zoom the minimap all the way out when the button is
-- right-clicked.
--
local function minimapZoomOut(self, button)
  -- Check the option is enabled
  if (NigglesDBC.settings.UiMiniMapZoom == true) then
    -- Check the right button was clicked and the 'Zoom Out' button is enabled
    if ((button == "RightButton") and
        (MinimapZoomOut:IsEnabled())) then
      MinimapZoomIn:Enable();
      MinimapZoomOut:Disable();
      PlaySound(SOUNDKIT.IG_MINIMAP_ZOOM_OUT);
      Minimap:SetZoom(0);
    end
  end

  return;
end

--
-- Function to toggle Equipment Manager Ignore for a item slot.
-- This function implements functionality from the function
-- 'PaperDollFrameItemFlyoutButton_OnClick' in PaperDollFrame.lua.
--
local function paperDollItemSlotClicked(self)
  -- Check the option is enabled
  if (NigglesDBC.settings.UiEMIgnoreToggle == true) then
    -- Check the correct modifier key is down and Equipment Manager is shown
    if ((IsAltKeyDown()) and
        ((PaperDollEquipmentManagerPane:IsShown() and
         (PaperDollEquipmentManagerPane.selectedSetName)) or
         (GearManagerDialogPopup:IsShown())) and
         (self.popoutButton:IsShown())) then
      -- Get the ID of the slot
      local slotId = self:GetID();

      -- Toggle the ignoring of the slot
      if (EquipmentManagerIsSlotIgnoredForSave(slotId)) then
        EquipmentManagerUnignoreSlotForSave(slotId);
        self.ignored = nil;
      else
        EquipmentManagerIgnoreSlotForSave(slotId);
        self.ignored = true;
      end

      -- Update the slot and the flyout menu
      PaperDollItemSlotButton_Update(self);
      EquipmentFlyout_Show(self);
      PaperDollEquipmentManagerPaneSaveSet:Enable();
    end
  end

  return;
end

--
-- Function to process clicking of Niggles Stack Split 'Maximum' button
--
stackSplitMaxOnClick = function(self, mouseButton)
  -- Hide the 'Split Stack' frame
  StackSplitFrame:Hide();

  -- Check the 'Split Stack' frame has an owner
  if (StackSplitFrame.owner) then
    -- Call the owner's 'Split Stack' function, with max stack size
    StackSplitFrame.owner.SplitStack(StackSplitFrame.owner,
      stackSplitMaxPurchasable);
  end

  return;
end

--
-- Function called when the 'OpenStackSplit' function is called. It stores the
-- maximum amount purchasable and displays the 'Maximum' frame.
--
local function stackSplitOpen(maxPurchasable, parent, anchor, anchorTo)
  -- Local Variables
  local owner;
  local level;

  -- Save the maximum purchasable
  stackSplitMaxPurchasable = maxPurchasable;

  -- Find the ultimate owner of the 'Stack Split' frame
  owner = StackSplitFrame.owner;
  while ((owner ~= nil) and
         (owner ~= MerchantFrame) and
         (owner ~= UIParent)) do
    owner = owner:GetParent();
  end

  -- Show/Hide the 'Stack Split Max' frame, depending on the owner
  if ((NigglesDBC.settings.UiMerchantBuyMax == true) and
      (owner == MerchantFrame) and
      (maxPurchasable > 1)) then
    -- Create the 'Stack Split Max' frame, if required
    if (stackSplitMax == nil) then
      -- Create the 'Stack Split Max' frame
      stackSplitMax = CreateFrame("Frame", nil, StackSplitFrame,
        "NigglesStackSplitMaxTemplate");
      stackSplitMax.button:SetScript("OnClick", stackSplitMaxOnClick);

      -- Adjust the level of the 'Stack Split' buttons
      level = stackSplitMax:GetFrameLevel()+1;
      StackSplitOkayButton:SetFrameLevel(level);
      StackSplitCancelButton:SetFrameLevel(level);
    end

    -- Initialise the 'Stack Split Max' frame and show it
    stackSplitMax.button:SetText(format("%s (%d)", MAXIMUM, maxPurchasable));
    stackSplitMax:Show();
  else
    -- Hide the 'Stack Split Max' frame, if it exists
    if (stackSplitMax ~= nil) then
      stackSplitMax:Hide();
    end
  end

  return;
end

-------------------------------------------------------------------------------
--                A  D  D  O  N     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to automatically loot certain items
--
function L.autoLoot()
  -- Local Variables
  local _;
  local numLootItems = GetNumLootItems();
  local isLocked;
  local itemLink;
  local itemType;
  local settings = NigglesDBC.settings;

  -- Check if the loot is from fishing
  if ((settings.AutoLootFish == true) and (IsFishingLoot())) then
    for slotIdx = numLootItems, 1, -1 do
      _, _, _, _, isLocked = GetLootSlotInfo(slotIdx);
      if (isLocked == false) then
        tinsert(autoLootSlots, slotIdx);
      end
    end
  else
    -- Search for an item to auto-loot
    for slotIdx = numLootItems, 1, -1 do
      -- Check if the slot is locked
      _, _, _, _, _, isLocked = GetLootSlotInfo(slotIdx);
      if (isLocked == false) then
        -- Get info about the loot in the current slot
        itemType = GetLootSlotType(slotIdx);
        itemLink = GetLootSlotLink(slotIdx);

        -- Work out if the slot should be looted
        if (L.archaeologyIsLoot(itemType, itemLink)) then
          if (settings.AutoLootArch) then
            autoLootSlots[#autoLootSlots+1] = slotIdx;
          end
        elseif (((itemType == LOOT_SLOT_MONEY) and
                 (settings.AutoLootMoney)) or
                ((itemType == LOOT_SLOT_CURRENCY) and
                 (settings.AutoLootCurrency))) then
          autoLootSlots[#autoLootSlots+1] = slotIdx;
        end
      end
    end
  end

  -- Check a slot was found that should be looted
  if (next(autoLootSlots)) then
    C_Timer.After(LOOTFRAME_AUTOLOOT_DELAY, autoLootNext);
  end

  return;
end

-------------------------------------------------------------------------------
--               G  L  O  B  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to switch to the first/last page of a frame that uses a page
-- concept when a 'Prev' or 'Next' button is right-clicked.
--
function NigglesFirstLastPage(self, mouseButton)
  -- Local Variables
  local onClickFunc = self:GetScript("OnClick");
  local iterations = 0;

  -- Check the option is enabled and the right button was clicked
  if ((NigglesDBC.settings.UiFirstLastPage == true) and
      (mouseButton == "RightButton") and
      (onClickFunc ~= nil)) then
    -- Call the button's 'OnClick' handler until it is disabled
    while ((self:IsEnabled()) and (iterations < 50)) do
      onClickFunc(self, mouseButton);
      iterations = iterations+1;
    end
  end

  return;
end

--
-- Function to hook into the standard UI
--
-- NOTE: This function should only ever be called once
--
function NigglesHookUi()
  -- Hook functions into the Zoom buttons on the Minimap
  MinimapZoomIn:HookScript("OnMouseUp", minimapZoomIn);
  MinimapZoomOut:HookScript("OnMouseUp", minimapZoomOut);

  -- Hook the clicking of any container item button
  hooksecurefunc("ContainerFrameItemButton_OnModifiedClick",
    bagItemClicked);
  hooksecurefunc("BankFrameItemButtonGeneric_OnModifiedClick",
    bankItemClicked);

  -- Hook a function into the equipment buttons
  hooksecurefunc("PaperDollItemSlotButton_OnModifiedClick",
    paperDollItemSlotClicked);

  -- Hook into the Game Time frame
  GameTimeFrame:HookScript("OnClick", NigglesCalendarEventsOnClick);

  -- Hook functions into the 'Stack Split' frame
  hooksecurefunc("OpenStackSplitFrame", stackSplitOpen);

  return;
end

--
-- Function to enable/disable features initialised as hooks into standard UI
--
function NigglesHookUiEnable()
  -- Enable/Disable the 'Stack Split Max' frame
  if (StackSplitFrame:IsShown()) then
    stackSplitOpen(stackSplitMaxPurchasable, StackSplitFrame.owner);
  end

  return;
end
